
import copy
from distutils.debug import DEBUG
from types import prepare_class

import numpy as np

# from PySide6 import QtGui, QtCore

# import GraphicsItemsCollection as gic
# import GraphicsItem
from foilmesh import connect
import logging
logger = logging.getLogger(__name__)

class SmoothAngleBased:
    """Mesh smoothing based on the paper:

    An Angular Method with Position Control for Block Mesh Squareness Improvement
    by Jin Yao, Douglas Stillman

    There are several errors in the paper.
    Nevertheless, the idea and algorithm description are formulated very clear.

    The errors are in the summation over all 12 angles, which in fact needs
    to be split into four for the alphas and eight for the betas.
    Therefore also the depicted derivatives are wrong as well as
    the inverse Hessian for the Newton optimization iterations.

    This class contains the corrected equations. 
    """

    def __init__(self, data, data_source='block'):
        
        # get MainWindow instance (overcomes handling parents)
        # self.mainwindow = QtCore.QCoreApplication.instance().mainwindow

        # data_source is one of 'block' or 'mesh'
        if data_source == 'block':
            self.block = data
            connect = connect.Connect(None)
            vertices = connect.getVertices(self.block)
            connectivity = connect.getConnectivity(self.block)
            self.mesh = vertices, connectivity
        if data_source == 'mesh':
            self.mesh = data

        lvc = self.makeLVC()
        self.stencils = self.make_stencil(lvc)

        self.drawlines = None

    def makeLVC(self):
        _, connectivity = self.mesh
        nodes = set([node for cell in connectivity for node in cell])

        self.lvc = dict()
        conn = np.array(connectivity)

        for node in nodes:
            cells, _ = np.where(conn == node)
            self.lvc.setdefault(node, []).append([conn[cell] for cell in cells])

        return self.lvc

    def make_stencil(self, lvc, verbose=False):     
        # lvc is a dictionary
        self.stencils = dict()
        for idx in range(len(lvc)):
            v, c = np.unique(lvc[idx], return_counts=True)
            vertices_quad = v[np.argwhere(c==1)]
            vertices_star = v[np.argwhere(c==2)]
    
            cells_with_common_edges = list()
            if len(vertices_star) != 4:
                continue
            
            for vertex in vertices_star:
                mask = np.isin(lvc[idx], [vertex[0], idx])
                mask1 = np.count_nonzero(mask, axis=1) == 2
                cells_with_common_edges.append(np.array(lvc[idx])[mask1])

            if idx == 6 and verbose:
            
                for cells in cells_with_common_edges:
                    mask2 = np.isin(cells, np.append(vertices_star.flatten(), idx))
    
            corresponding_corners = list()
            for cells in cells_with_common_edges:
                mask2 = np.isin(cells, np.append(vertices_star.flatten(), idx))
                corresponding_corners.append(cells[~mask2])
    
            self.stencils[idx] = corresponding_corners
    
        return self.stencils

    def make_cardinals(self, vertices):

        cardinals = dict()
        
        for stencil in self.stencils:
            s = self.stencils[stencil]

            D = [vertices[s[2][0]][0], vertices[s[2][0]][1]]
            EE = [vertices[s[0][0]][0], vertices[s[0][0]][1]]
            F = [vertices[s[1][1]][0], vertices[s[1][1]][1]]
            G = [vertices[s[0][1]][0], vertices[s[0][1]][1]]

            S = (0.5 * (D[0] + EE[0]), 0.5 * (D[1] + EE[1]))
            W = (0.5 * (D[0] + G[0]), 0.5 * (D[1] + G[1]))
            E = (0.5 * (EE[0] + F[0]), 0.5 * (EE[1] + F[1]))
            N = (0.5 * (G[0] + F[0]), 0.5 * (G[1] + F[1]))
            cardinals[stencil] = (S, W, E, N, D, EE, F, G)
        return cardinals

    # def draw_cardinal(self, S, W, E, N, D, EE, F, G):

    #     self.drawlines = list()

    #     gc = gic.GraphicsCollection()

    #     points = [QtCore.QPointF(x, y) for x, y in [S, N]]
    #     gc.Polyline(QtGui.QPolygonF(points), '')
    #     gc.pen.setColor(QtGui.QColor(255, 0, 0, 255))
    #     gc.pen.setWidthF(3.0)
    #     gc.pen.setCosmetic(True)
    #     gc.brush.setStyle(QtCore.Qt.NoBrush)
    #     meshline = GraphicsItem.GraphicsItem(gc)
    #     self.drawlines.append(meshline)

    #     gc = gic.GraphicsCollection()
    #     points = [QtCore.QPointF(x, y) for x, y in [E, W]]
    #     gc.Polyline(QtGui.QPolygonF(points), '')
    #     gc.pen.setColor(QtGui.QColor(0, 255, 0, 255))
    #     gc.pen.setWidthF(3.0)
    #     gc.pen.setCosmetic(True)
    #     gc.brush.setStyle(QtCore.Qt.NoBrush)
    #     meshline = GraphicsItem.GraphicsItem(gc)
    #     self.drawlines.append(meshline)

    #     gc = gic.GraphicsCollection()
    #     points = [QtCore.QPointF(x, y) for x, y in [D, EE, F, G, D]]
    #     gc.Polyline(QtGui.QPolygonF(points), '')
    #     gc.pen.setColor(QtGui.QColor(0, 0, 255, 255))
    #     gc.pen.setWidthF(5.0)
    #     gc.pen.setCosmetic(True)
    #     gc.brush.setStyle(QtCore.Qt.NoBrush)
    #     meshline = GraphicsItem.GraphicsItem(gc)
    #     self.drawlines.append(meshline)

    #     '''
    #     gc = gic.GraphicsCollection()
    #     points = [QtCore.QPointF(x, y) for x, y in [EE, G]]
    #     gc.Polyline(QtGui.QPolygonF(points), '')
    #     gc.pen.setColor(QtGui.QColor(0, 255, 255, 255))
    #     gc.pen.setWidthF(9.0)
    #     gc.pen.setCosmetic(True)
    #     gc.brush.setStyle(QtCore.Qt.NoBrush)
    #     meshline = GraphicsItem.GraphicsItem(gc)
    #     self.drawlines.append(meshline)

    #     gc = gic.GraphicsCollection()
    #     points = [QtCore.QPointF(x, y) for x, y in [D, G]]
    #     gc.Polyline(QtGui.QPolygonF(points), '')
    #     gc.pen.setColor(QtGui.QColor(255, 0, 255, 255))
    #     gc.pen.setWidthF(9.0)
    #     gc.pen.setCosmetic(True)
    #     gc.brush.setStyle(QtCore.Qt.NoBrush)
    #     meshline = GraphicsItem.GraphicsItem(gc)
    #     self.drawlines.append(meshline)

    #     gc = gic.GraphicsCollection()
    #     points = [QtCore.QPointF(x, y) for x, y in [D, F]]
    #     gc.Polyline(QtGui.QPolygonF(points), '')
    #     gc.pen.setColor(QtGui.QColor(0, 0, 0, 255))
    #     gc.pen.setWidthF(9.0)
    #     gc.pen.setCosmetic(True)
    #     gc.brush.setStyle(QtCore.Qt.NoBrush)
    #     meshline = GraphicsItem.GraphicsItem(gc)
    #     self.drawlines.append(meshline)
    #     '''

    #     # self.mainwindow.scene.createItemGroup(self.drawlines)

    def smooth(self, iterations=20, tolerance=1.e-4, verbose=False):

        # iterations=1

        vertices, _ = self.mesh
    
        cardinals = self.make_cardinals(vertices)

        smoothed_vertices = copy.deepcopy(vertices)
        smoothed_vertices_old = copy.deepcopy(vertices)
    
        corner = False
        omega = 1
        if corner:
            omega = 0

        # loop until convergence
        iteration = 0
        while iteration < iterations:
            iteration += 1

            # loop over all stencils (for vertices to be smoothed)
            for ic, cardinal in enumerate(cardinals):

                (x, y) = smoothed_vertices[cardinal]
                (xold, yold) = smoothed_vertices_old[cardinal]

                S, W, E, N, D, EE, F, G = cardinals[cardinal]

                DEBUG = False
                if DEBUG and ic == 143:
                    self.draw_cardinal(S, W, E, N, D, EE, F, G)
                    # print('ic, cardinal', ic, cardinal)
                    # print('Stencil', self.stencils[cardinal])

                # calculate position control
                NS = np.linalg.norm( (S[0] - N[0], S[1] - N[1]) )
                WE = np.linalg.norm( (E[0] - W[0], E[1] - W[1]) )
                sigma = np.max((NS/WE, WE/NS))

                # angles alpha
                a1 = np.array([S[0], E[0], N[0], W[0]])
                a2 = np.array([E[0], N[0], W[0], S[0]])
                b1 = np.array([S[1], E[1], N[1], W[1]])
                b2 = np.array([E[1], N[1], W[1], S[1]])

                # angles beta
                c1 = np.array([S[0], S[0], E[0], E[0], N[0], N[0], W[0], W[0]])
                c2 = np.array([D[0], EE[0], EE[0], F[0], F[0], G[0], G[0], D[0]])
                d1 = np.array([S[1], S[1], E[1], E[1], N[1], N[1], W[1], W[1]])
                d2 = np.array([D[1], EE[1], EE[1], F[1], F[1], G[1], G[1], D[1]])

                # derivatives of alpha contributions (including position control)
                ca = np.sum(omega / ( (a1**2 + b1**2 - 2*a1*xold + xold**2 - 2*b1*yold + yold**2) * \
                                      (a2**2 + b2**2 - 2*a2*xold + xold**2 - 2*b2*yold + yold**2) + 1.e-9))                
                dTdx_alpha = np.sum(-(a1*a2 + b1*b2 - a1*x - a2*x + x**2 - b1*y - b2*y + y**2) * \
                                (a1 + a2 - 2.*x) - (a1 + a1 + a1 + a1 - 4.*x) * sigma)
                dTdy_alpha = np.sum(-(a1*a2 + b1*b2 - a1*x - a2*x + x**2 - b1*y - b2*y + y**2) * \
                                (b1 + b2 - 2*y) - (b1 + b1 + b1 + b1 - 4*y)*sigma)
                d2Tdx2_alpha = np.sum((a1 + a2 - 2*x)**2 + 2*a1*a2 + 2*b1*b2 - 2*a1*x - 2*a2*x + 2*x**2 - \
                                2*b1*y - 2*b2*y + 2*y**2 + 4*sigma)
                d2Tdy2_alpha = np.sum(2*a1*a2 + (b1 + b2 - 2*y)**2 + 2*b1*b2 - 2*a1*x - 2*a2*x + 2*x**2 - \
                                2*b1*y - 2*b2*y + 2*y**2 + 4*sigma)
                d2Tdxdy_alpha = np.sum((a1 + a2 - 2*x)*(b1 + b2 - 2*y))

                # derivatives of beta contributions (including position control)
                cb = np.sum(omega / ((c1**2 - 2*c1*c2 + c2**2 + d1**2 - 2*d1*d2 + d2**2) * \
                                        (c1**2 + d1**2 - 2*c1*xold + xold**2 - 2*d1*yold + yold**2) + 1.e-9))
                dTdx_beta = np.sum(-(c1**2 - c1*c2 + d1**2 - d1*d2 - c1*x + c2*x - d1*y + d2*y) * (c1 - c2) -\
                                        (a1[0] + a1[1] + a1[2] + a1[3] - 4.*x)*sigma)
                dTdy_beta = np.sum(-(c1**2 - c1*c2 + d1**2 - d1*d2 - c1*x + c2*x - d1*y + d2*y) * (d1 - d2) -\
                                        (b1[0] + b1[1] + b1[2] + b1[3] - 4*y)*sigma)
                d2Tdx2_beta = np.sum((c1 - c2)**2 + 4*sigma)
                d2Tdy2_beta = np.sum((d1 - d2)**2 + 4*sigma)
                d2Tdxdy_beta = np.sum((c1 - c2)*(d1 - d2))

                # compile derivatives of all contributions
                dTdx = ca * dTdx_alpha + cb * dTdx_beta
                dTdy = ca * dTdy_alpha + cb * dTdy_beta
                d2Tdx2 = ca * d2Tdx2_alpha + cb * d2Tdx2_beta
                d2Tdy2 = ca * d2Tdy2_alpha + cb * d2Tdy2_beta
                d2Tdxdy = ca * d2Tdxdy_alpha + cb * d2Tdxdy_beta

                # Newton iteration for optimization
                xnew = x - [d2Tdy2 * dTdx - d2Tdxdy * dTdy] / (d2Tdx2 * d2Tdy2 - (d2Tdxdy)**2)
                ynew = y - [d2Tdx2 * dTdy - d2Tdxdy * dTdx] / (d2Tdx2 * d2Tdy2 - (d2Tdxdy)**2)

                smoothed_vertices[cardinal] = (xnew[0], ynew[0])
                smoothed_vertices_old[cardinal] = (x, y)

                tol = np.linalg.norm((xnew[0] - x, ynew[0] - y))

            if verbose:
                logger.info(f'Iteration={iteration:3d}, residual={tol:.3e}')

            if tol < tolerance:
                break

            # update current cardinals for next iteration
            cardinals = self.make_cardinals(smoothed_vertices)

        if self.drawlines:
            self.mainwindow.scene.createItemGroup(self.drawlines)

        return smoothed_vertices
    
    def mapToUlines(self, smoothed_vertices):

        self.new_ulines = list()

        j = -1
        for uline in self.block.getULines():
            new_uline = list()
            for i in range(len(uline)):
                j += 1
                new_uline.append(smoothed_vertices[j])

            self.new_ulines.append(new_uline)

        return self.new_ulines



from functools import reduce

import numpy as np

from ..__about__ import __version__
from .._common import info, join_strings, replace_space, warn
from .._exceptions import ReadError, WriteError
from .._files import open_file
from .._mesh import Mesh
from .._vtk_common import (
    Info,
    meshio_to_vtk_order,
    meshio_to_vtk_type,
    vtk_cells_from_data,
)

# VTK 5.1 data types
vtk_to_numpy_dtype_name = {
    "float": "float32",
    "double": "float64",
    "int": "int",
    "vtktypeint8": "int8",
    "vtktypeint16": "int16",
    "vtktypeint32": "int32",
    "vtktypeint64": "int64",
    "vtktypeuint8": "uint8",
    "vtktypeuint16": "uint16",
    "vtktypeuint32": "uint32",
    "vtktypeuint64": "uint64",
}
numpy_to_vtk_dtype = {v: k for k, v in vtk_to_numpy_dtype_name.items()}

# supported vtk dataset types
vtk_dataset_types = [
    "UNSTRUCTURED_GRID",
    "STRUCTURED_POINTS",
    "STRUCTURED_GRID",
    "RECTILINEAR_GRID",
]
# additional infos per dataset type
vtk_dataset_infos = {
    "UNSTRUCTURED_GRID": [],
    "STRUCTURED_POINTS": [
        "DIMENSIONS",
        "ORIGIN",
        "SPACING",
        "ASPECT_RATIO",  # alternative for SPACING in version 1.0 and 2.0
    ],
    "STRUCTURED_GRID": ["DIMENSIONS"],
    "RECTILINEAR_GRID": [
        "DIMENSIONS",
        "X_COORDINATES",
        "Y_COORDINATES",
        "Z_COORDINATES",
    ],
}

# all main sections in vtk
vtk_sections = [
    "METADATA",
    "DATASET",
    "POINTS",
    "CELLS",
    "CELL_TYPES",
    "POINT_DATA",
    "CELL_DATA",
    "LOOKUP_TABLE",
    "COLOR_SCALARS",
]


def read(filename):
    with open_file(filename, "rb") as f:
        out = read_buffer(f)
    return out


def read_buffer(f):
    # initialize output data
    info = Info()

    # skip title comment
    f.readline()

    data_type = f.readline().decode().strip().upper()
    if data_type not in ["ASCII", "BINARY"]:
        raise ReadError(f"Unknown VTK data type '{data_type}'.")
    info.is_ascii = data_type == "ASCII"

    while True:
        line = f.readline().decode()
        if not line:
            # EOF
            break

        line = line.strip()
        if len(line) == 0:
            continue

        info.split = line.split()
        info.section = info.split[0].upper()

        if info.section in vtk_sections:
            _read_section(f, info)
        else:
            _read_subsection(f, info)

    _check_mesh(info)

    cells, cell_data = vtk_cells_from_data(
        info.connectivity, info.offsets, info.types, info.cell_data_raw
    )

    return Mesh(
        info.points,
        cells,
        point_data=info.point_data,
        cell_data=cell_data,
        field_data=info.field_data,
    )


def _read_section(f, info):
    if info.section == "METADATA":
        _skip_meta(f)

    elif info.section == "DATASET":
        info.active = "DATASET"
        info.dataset["type"] = info.split[1].upper()
        if info.dataset["type"] not in vtk_dataset_types:
            legal = ", ".join(vtk_dataset_types)
            raise ReadError(
                f"Only VTK '{legal}' supported (not {info.dataset['type']})."
            )

    elif info.section == "POINTS":
        info.active = "POINTS"
        info.num_points = int(info.split[1])
        data_type = info.split[2].lower()
        info.points = _read_points(f, data_type, info.is_ascii, info.num_points)

    elif info.section == "CELLS":
        info.active = "CELLS"
        try:
            line = f.readline().decode()
        except UnicodeDecodeError:
            line = ""

        assert line.startswith("OFFSETS")
        # vtk DataFile Version 5.1 - appearing in Paraview 5.8.1 outputs
        # No specification found for this file format.
        # See the question on ParaView Discourse Forum:
        # https://discourse.paraview.org/t/specification-of-vtk-datafile-version-5-1/5127
        info.num_offsets = int(info.split[1])
        info.num_items = int(info.split[2])
        dtype = np.dtype(vtk_to_numpy_dtype_name[line.split()[1]])
        offsets = _read_int_data(f, info.is_ascii, info.num_offsets, dtype)

        line = f.readline().decode()
        assert line.startswith("CONNECTIVITY")
        dtype = np.dtype(vtk_to_numpy_dtype_name[line.split()[1]])
        connectivity = _read_int_data(f, info.is_ascii, info.num_items, dtype)
        info.connectivity = connectivity
        assert offsets[0] == 0
        assert offsets[-1] == len(connectivity)
        info.offsets = offsets[1:]

    elif info.section == "CELL_TYPES":
        info.active = "CELL_TYPES"
        info.num_items = int(info.split[1])
        info.types = _read_cell_types(f, info.is_ascii, info.num_items)

    elif info.section == "POINT_DATA":
        info.active = "POINT_DATA"
        info.num_items = int(info.split[1])

    elif info.section == "CELL_DATA":
        info.active = "CELL_DATA"
        info.num_items = int(info.split[1])

    elif info.section == "LOOKUP_TABLE":
        info.num_items = int(info.split[2])
        np.fromfile(f, count=info.num_items * 4, sep=" ", dtype=float)
        # rgba = data.reshape((info.num_items, 4))

    elif info.section == "COLOR_SCALARS":
        nValues = int(info.split[2])
        # re-use num_items from active POINT/CELL_DATA
        num_items = info.num_items
        dtype = np.ubyte
        if info.is_ascii:
            dtype = float
        np.fromfile(f, count=num_items * nValues, dtype=dtype)


def _read_subsection(f, info):
    if info.active == "POINT_DATA":
        d = info.point_data
    elif info.active == "CELL_DATA":
        d = info.cell_data_raw
    elif info.active == "DATASET":
        d = info.dataset
    else:
        d = info.field_data

    if info.section in vtk_dataset_infos[info.dataset["type"]]:
        if info.section[1:] == "_COORDINATES":
            info.num_points = int(info.split[1])
            data_type = info.split[2].lower()
            d[info.section] = _read_coords(f, data_type, info.is_ascii, info.num_points)
        else:
            if info.section == "DIMENSIONS":
                d[info.section] = list(map(int, info.split[1:]))
            else:
                d[info.section] = list(map(float, info.split[1:]))
            if len(d[info.section]) != 3:
                raise ReadError(
                    f"Wrong number of info in section '{info.section}'. "
                    f"Need 3, got {len(d[info.section])}."
                )
    elif info.section == "SCALARS":
        d.update(_read_scalar_field(f, info.num_items, info.split, info.is_ascii))
    elif info.section == "VECTORS":
        d.update(_read_field(f, info.num_items, info.split, [3], info.is_ascii))
    elif info.section == "TENSORS":
        d.update(_read_field(f, info.num_items, info.split, [3, 3], info.is_ascii))
    elif info.section == "FIELD":
        d.update(_read_fields(f, int(info.split[2]), info.is_ascii))
    else:
        raise ReadError(f"Unknown section '{info.section}'.")


def _check_mesh(info):
    if info.dataset["type"] == "UNSTRUCTURED_GRID":
        if info.connectivity is None:
            raise ReadError("Required section CELLS not found.")
        if info.types is None:
            raise ReadError("Required section CELL_TYPES not found.")
    elif info.dataset["type"] == "STRUCTURED_POINTS":
        dim = info.dataset["DIMENSIONS"]
        ori = info.dataset["ORIGIN"]
        spa = (
            info.dataset["SPACING"]
            if "SPACING" in info.dataset
            else info.dataset["ASPECT_RATIO"]
        )
        axis = [
            np.linspace(ori[i], ori[i] + (dim[i] - 1.0) * spa[i], dim[i])
            for i in range(3)
        ]
        info.points = _generate_points(axis)
        info.connectivity, info.types = _generate_cells(dim=info.dataset["DIMENSIONS"])
    elif info.dataset["type"] == "RECTILINEAR_GRID":
        axis = [
            info.dataset["X_COORDINATES"],
            info.dataset["Y_COORDINATES"],
            info.dataset["Z_COORDINATES"],
        ]
        info.points = _generate_points(axis)
        info.connectivity, info.types = _generate_cells(dim=info.dataset["DIMENSIONS"])
    elif info.dataset["type"] == "STRUCTURED_GRID":
        info.connectivity, info.types = _generate_cells(dim=info.dataset["DIMENSIONS"])


def _generate_cells(dim):
    ele_dim = [d - 1 for d in dim if d > 1]
    # TODO use math.prod when requiring Python 3.8+? this would save the int conversion
    # <https://github.com/microsoft/pyright/issues/1226>
    ele_no = int(np.prod(ele_dim))
    spatial_dim = len(ele_dim)

    if spatial_dim == 1:
        # cells are lines in 1D
        cells = np.empty((ele_no, 3), dtype=int)
        cells[:, 0] = 2
        cells[:, 1] = np.arange(ele_no, dtype=int)
        cells[:, 2] = cells[:, 1] + 1
        cell_types = np.full(ele_no, 3, dtype=int)
    elif spatial_dim == 2:
        # cells are quad in 2D
        cells = np.empty((ele_no, 5), dtype=int)
        cells[:, 0] = 4
        cells[:, 1] = np.arange(0, ele_no, dtype=int)
        cells[:, 1] += np.arange(0, ele_no, dtype=int) // ele_dim[0]
        cells[:, 2] = cells[:, 1] + 1
        cells[:, 3] = cells[:, 1] + 2 + ele_dim[0]
        cells[:, 4] = cells[:, 3] - 1
        cell_types = np.full(ele_no, 9, dtype=int)
    else:
        # cells are hex in 3D
        cells = np.empty((ele_no, 9), dtype=int)
        cells[:, 0] = 8
        cells[:, 1] = np.arange(ele_no)
        cells[:, 1] += (ele_dim[0] + ele_dim[1] + 1) * (
            np.arange(ele_no) // (ele_dim[0] * ele_dim[1])
        )
        cells[:, 1] += (np.arange(ele_no) % (ele_dim[0] * ele_dim[1])) // ele_dim[0]
        cells[:, 2] = cells[:, 1] + 1
        cells[:, 3] = cells[:, 1] + 2 + ele_dim[0]
        cells[:, 4] = cells[:, 3] - 1
        cells[:, 5] = cells[:, 1] + (1 + ele_dim[0]) * (1 + ele_dim[1])
        cells[:, 6] = cells[:, 5] + 1
        cells[:, 7] = cells[:, 5] + 2 + ele_dim[0]
        cells[:, 8] = cells[:, 7] - 1
        cell_types = np.full(ele_no, 12, dtype=int)

    return cells.reshape(-1), cell_types


def _generate_points(axis):
    x_dim = len(axis[0])
    y_dim = len(axis[1])
    z_dim = len(axis[2])
    pnt_no = x_dim * y_dim * z_dim
    x_id, y_id, z_id = np.mgrid[0:x_dim, 0:y_dim, 0:z_dim]
    points = np.empty((pnt_no, 3), dtype=axis[0].dtype)
    # VTK sorts points and cells in Fortran order
    points[:, 0] = axis[0][x_id.reshape(-1, order="F")]
    points[:, 1] = axis[1][y_id.reshape(-1, order="F")]
    points[:, 2] = axis[2][z_id.reshape(-1, order="F")]
    return points


def _read_coords(f, data_type, is_ascii, num_points):
    dtype = np.dtype(vtk_to_numpy_dtype_name[data_type])
    if is_ascii:
        coords = np.fromfile(f, count=num_points, sep=" ", dtype=dtype)
    else:
        # Binary data is big endian, see
        # <https://vtk.org/Wiki/VTK/Writing_VTK_files_using_python#.22legacy.22>.
        dtype = dtype.newbyteorder(">")
        coords = np.fromfile(f, count=num_points, dtype=dtype)
        line = f.readline().decode()
        if line != "\n":
            raise ReadError()
    return coords


def _read_points(f, data_type, is_ascii, num_points):
    dtype = np.dtype(vtk_to_numpy_dtype_name[data_type])
    if is_ascii:
        points = np.fromfile(f, count=num_points * 3, sep=" ", dtype=dtype)
    else:
        # Binary data is big endian, see
        # <https://vtk.org/Wiki/VTK/Writing_VTK_files_using_python#.22legacy.22>.
        dtype = dtype.newbyteorder(">")
        points = np.fromfile(f, count=num_points * 3, dtype=dtype)
        line = f.readline().decode()
        if line != "\n":
            raise ReadError()
    return points.reshape((num_points, 3))


def _read_int_data(f, is_ascii, num_items, dtype):
    if is_ascii:
        c = np.fromfile(f, count=num_items, sep=" ", dtype=dtype)
    else:
        dtype = dtype.newbyteorder(">")
        c = np.fromfile(f, count=num_items, dtype=dtype)
        line = f.readline().decode()
        if line != "\n":
            raise ReadError("Expected newline")
    return c


def _read_cell_types(f, is_ascii, num_items):
    if is_ascii:
        ct = np.fromfile(f, count=int(num_items), sep=" ", dtype=int)
    else:
        # binary
        ct = np.fromfile(f, count=int(num_items), dtype=">i4")
        line = f.readline().decode()
        # Sometimes, there's no newline at the end
        if line.strip() != "":
            raise ReadError()
    return ct


def _read_scalar_field(f, num_data, split, is_ascii):
    data_name = split[1]
    data_type = split[2].lower()
    try:
        num_comp = int(split[3])
    except IndexError:
        num_comp = 1

    # The standard says:
    # > The parameter numComp must range between (1,4) inclusive; [...]
    if not (0 < num_comp < 5):
        raise ReadError("The parameter numComp must range between (1,4) inclusive")

    dtype = np.dtype(vtk_to_numpy_dtype_name[data_type])
    lt, _ = f.readline().decode().split()
    if lt.upper() != "LOOKUP_TABLE":
        raise ReadError()

    if is_ascii:
        data = np.fromfile(f, count=num_data * num_comp, sep=" ", dtype=dtype)
    else:
        # Binary data is big endian, see
        # <https://vtk.org/Wiki/VTK/Writing_VTK_files_using_python#.22legacy.22>.
        dtype = dtype.newbyteorder(">")
        data = np.fromfile(f, count=num_data * num_comp, dtype=dtype)
        line = f.readline().decode()
        if line != "\n":
            raise ReadError()

    data = data.reshape(-1, num_comp)
    return {data_name: data}


def _read_field(f, num_data, split, shape, is_ascii):
    data_name = split[1]
    data_type = split[2].lower()

    dtype = np.dtype(vtk_to_numpy_dtype_name[data_type])
    # prod()
    # <https://stackoverflow.com/q/2104782/353337>
    k = reduce((lambda x, y: x * y), shape)

    if is_ascii:
        data = np.fromfile(f, count=k * num_data, sep=" ", dtype=dtype)
    else:
        # Binary data is big endian, see
        # <https://vtk.org/Wiki/VTK/Writing_VTK_files_using_python#.22legacy.22>.
        dtype = dtype.newbyteorder(">")
        data = np.fromfile(f, count=k * num_data, dtype=dtype)
        line = f.readline().decode()
        if line != "\n":
            raise ReadError()

    data = data.reshape(-1, *shape)
    return {data_name: data}


def _read_fields(f, num_fields, is_ascii):
    data = {}
    for _ in range(num_fields):
        line = f.readline().decode().split()
        if line[0] == "METADATA":
            _skip_meta(f)
            name, shape0, shape1, data_type = f.readline().decode().split()
        else:
            name, shape0, shape1, data_type = line

        shape0 = int(shape0)
        shape1 = int(shape1)
        dtype = np.dtype(vtk_to_numpy_dtype_name[data_type.lower()])

        if is_ascii:
            dat = np.fromfile(f, count=shape0 * shape1, sep=" ", dtype=dtype)
        else:
            # Binary data is big endian, see
            # <https://vtk.org/Wiki/VTK/Writing_VTK_files_using_python#.22legacy.22>.
            dtype = dtype.newbyteorder(">")
            dat = np.fromfile(f, count=shape0 * shape1, dtype=dtype)
            line = f.readline().decode()
            if line != "\n":
                raise ReadError()

        if shape0 != 1:
            dat = dat.reshape((shape1, shape0))

        data[name] = dat

    return data


def _skip_meta(f):
    # skip possible metadata
    # https://vtk.org/doc/nightly/html/IOLegacyInformationFormat.html
    while True:
        line = f.readline().decode().strip()
        if not line:
            # end of metadata is a blank line
            break


def _pad(array):
    return np.pad(array, ((0, 0), (0, 1)), "constant")


def write(filename, mesh, binary=True):
    if mesh.points.shape[1] == 2:
        warn(
            "VTK requires 3D points, but 2D points given. "
            "Appending 0 third component."
        )
        points = _pad(mesh.points)
    else:
        points = mesh.points

    if mesh.point_data:
        for name, values in mesh.point_data.items():
            if len(values.shape) == 2 and values.shape[1] == 2:
                warn(
                    "VTK requires 3D vectors, but 2D vectors given. "
                    f"Appending 0 third component to {name}."
                )
                mesh.point_data[name] = _pad(values)

    for name, data in mesh.cell_data.items():
        for k, values in enumerate(data):
            if len(values.shape) == 2 and values.shape[1] == 2:
                warn(
                    "VTK requires 3D vectors, but 2D vectors given. "
                    f"Appending 0 third component to {name}."
                )
                data[k] = _pad(data[k])

    if not binary:
        warn("VTK ASCII files are only meant for debugging.")

    if mesh.point_sets:
        info(
            "VTK format cannot write point_sets. Converting them to point_data...",
            highlight=False,
        )
        key, _ = join_strings(list(mesh.point_sets.keys()))
        key, _ = replace_space(key)
        mesh.point_sets_to_data(key)

    if mesh.cell_sets:
        info(
            "VTK format cannot write cell_sets. Converting them to cell_data...",
            highlight=False,
        )
        key, _ = join_strings(list(mesh.cell_sets.keys()))
        key, _ = replace_space(key)
        mesh.cell_sets_to_data(key)

    with open_file(filename, "wb") as f:
        f.write(b"# vtk DataFile Version 5.1\n")
        f.write(f"written by meshio v{__version__}\n".encode())
        f.write(("BINARY\n" if binary else "ASCII\n").encode())
        f.write(b"DATASET UNSTRUCTURED_GRID\n")

        # write points and cells
        _write_points(f, points, binary)
        _write_cells(f, mesh.cells, binary)

        # write point data
        if mesh.point_data:
            num_points = mesh.points.shape[0]
            f.write(f"POINT_DATA {num_points}\n".encode())
            _write_field_data(f, mesh.point_data, binary)

        # write cell data
        if mesh.cell_data:
            total_num_cells = sum(len(c.data) for c in mesh.cells)
            f.write(f"CELL_DATA {total_num_cells}\n".encode())
            _write_field_data(f, mesh.cell_data, binary)


def _write_points(f, points, binary):
    dtype = numpy_to_vtk_dtype[points.dtype.name]
    f.write(f"POINTS {len(points)} {dtype}\n".encode())

    if binary:
        # Binary data must be big endian, see
        # <https://vtk.org/Wiki/VTK/Writing_VTK_files_using_python#.22legacy.22>.
        # if points.dtype.byteorder == "<" or (
        #     points.dtype.byteorder == "=" and sys.byteorder == "little"
        # ):
        #     logging.warn("Converting to new byte order")
        points.astype(points.dtype.newbyteorder(">")).tofile(f, sep="")
    else:
        # ascii
        points.tofile(f, sep=" ")
    f.write(b"\n")


def _write_cells(f, cells, binary):
    total_num_cells = sum(len(c.data) for c in cells)
    total_num_idx = sum(c.data.size for c in cells)
    f.write(f"CELLS {total_num_cells + 1} {total_num_idx}\n".encode())

    # offsets
    offsets = [[0]]
    k = 0
    for cell_block in cells:
        m, n = cell_block.data.shape
        offsets.append(np.arange(k + n, k + (m + 1) * n, n))
        k = offsets[-1][-1]
    offsets = np.concatenate(offsets)

    if binary:
        f.write(b"OFFSETS vtktypeint64\n")
        # force big-endian and int64
        offsets.astype(">i8").tofile(f, sep="")
        f.write(b"\n")

        f.write(b"CONNECTIVITY vtktypeint64\n")
        for cell_block in cells:
            d = cell_block.data
            cell_idx = meshio_to_vtk_order(cell_block.type)
            if cell_idx is not None:
                d = d[:, cell_idx]
            # force big-endian and int64
            d.astype(">i8").tofile(f, sep="")
        f.write(b"\n")
    else:
        # ascii
        f.write(b"OFFSETS vtktypeint64\n")
        offsets.tofile(f, sep="\n")
        f.write(b"\n")

        f.write(b"CONNECTIVITY vtktypeint64\n")
        for cell_block in cells:
            d = cell_block.data
            cell_idx = meshio_to_vtk_order(cell_block.type)
            if cell_idx is not None:
                d = d[:, cell_idx]
            d.tofile(f, sep="\n")
            f.write(b"\n")

    # write cell types
    f.write(f"CELL_TYPES {total_num_cells}\n".encode())
    if binary:
        for c in cells:
            vtk_type = meshio_to_vtk_type[c.type]
            np.full(len(c.data), vtk_type, dtype=np.dtype(">i4")).tofile(f, sep="")
        f.write(b"\n")
    else:
        # ascii
        for c in cells:
            vtk_type = meshio_to_vtk_type[c.type]
            np.full(len(c.data), vtk_type).tofile(f, sep="\n")
            f.write(b"\n")


def _write_field_data(f, data, binary):
    f.write((f"FIELD FieldData {len(data)}\n").encode())
    for name, values in data.items():
        if isinstance(values, list):
            values = np.concatenate(values)
        if len(values.shape) == 1:
            num_tuples = values.shape[0]
            num_components = 1
        else:
            num_tuples = values.shape[0]
            num_components = values.shape[1]

        if " " in name:
            raise WriteError(f"VTK doesn't support spaces in field names ('{name}').")

        vtk_dtype = numpy_to_vtk_dtype[values.dtype.name]
        f.write(f"{name} {num_components} {num_tuples} {vtk_dtype}\n".encode())
        if binary:
            values.astype(values.dtype.newbyteorder(">")).tofile(f, sep="")
        else:
            # ascii
            values.tofile(f, sep=" ")
            # np.savetxt(f, points)
        f.write(b"\n")



"""
I/O for Ansys's msh format.

<https://romeo.univ-reims.fr/documents/fluent/tgrid/ug/appb.pdf>
"""
import re

import numpy as np

from ..__about__ import __version__
from .._common import warn
from .._exceptions import ReadError, WriteError
from .._files import open_file
from .._helpers import register_format
from .._mesh import Mesh


def _skip_to(f, char):
    c = None
    while c != char:
        c = f.read(1).decode()


def _skip_close(f, num_open_brackets):
    while num_open_brackets > 0:
        char = f.read(1).decode()
        if char == "(":
            num_open_brackets += 1
        elif char == ")":
            num_open_brackets -= 1


def _read_points(f, line, first_point_index_overall, last_point_index):
    # If the line is self-contained, it is merely a declaration
    # of the total number of points.
    if line.count("(") == line.count(")"):
        return None, None, None

    # (3010 (zone-id first-index last-index type ND)
    out = re.match("\\s*\\(\\s*(|20|30)10\\s*\\(([^\\)]*)\\).*", line)
    assert out is not None
    a = [int(num, 16) for num in out.group(2).split()]

    if len(a) <= 4:
        raise ReadError()

    first_point_index = a[1]
    # store the very first point index
    if first_point_index_overall is None:
        first_point_index_overall = first_point_index
    # make sure that point arrays are subsequent
    if last_point_index is not None:
        if last_point_index + 1 != first_point_index:
            raise ReadError()
    last_point_index = a[2]
    num_points = last_point_index - first_point_index + 1
    dim = a[4]

    # Skip ahead to the byte that opens the data block (might
    # be the current line already).
    last_char = line.strip()[-1]
    while last_char != "(":
        last_char = f.read(1).decode()

    if out.group(1) == "":
        # ASCII data
        pts = np.empty((num_points, dim))
        for k in range(num_points):
            # skip ahead to the first line with data
            line = ""
            while line.strip() == "":
                line = f.readline().decode()
            dat = line.split()
            if len(dat) != dim:
                raise ReadError()
            for d in range(dim):
                pts[k][d] = float(dat[d])
    else:
        # binary data
        if out.group(1) == "20":
            dtype = np.float32
        else:
            if out.group(1) != "30":
                ReadError(f"Expected keys '20' or '30', got {out.group(1)}.")
            dtype = np.float64
        # read point data
        pts = np.fromfile(f, count=dim * num_points, dtype=dtype).reshape(
            (num_points, dim)
        )

    # make sure that the data set is properly closed
    _skip_close(f, 2)
    return pts, first_point_index_overall, last_point_index


def _read_cells(f, line):
    # If the line is self-contained, it is merely a declaration of the total number of
    # points.
    if line.count("(") == line.count(")"):
        return None, None

    out = re.match("\\s*\\(\\s*(|20|30)12\\s*\\(([^\\)]+)\\).*", line)
    assert out is not None
    a = [int(num, 16) for num in out.group(2).split()]
    if len(a) <= 4:
        raise ReadError()
    first_index = a[1]
    last_index = a[2]
    num_cells = last_index - first_index + 1
    zone_type = a[3]
    element_type = a[4]

    if zone_type == 0:
        # dead zone
        return None, None

    key, num_nodes_per_cell = {
        0: ("mixed", None),
        1: ("triangle", 3),
        2: ("tetra", 4),
        3: ("quad", 4),
        4: ("hexahedron", 8),
        5: ("pyramid", 5),
        6: ("wedge", 6),
    }[element_type]

    # Skip to the opening `(` and make sure that there's no non-whitespace character
    # between the last closing bracket and the `(`.
    if line.strip()[-1] != "(":
        c = None
        while True:
            c = f.read(1).decode()
            if c == "(":
                break
            if not re.match("\\s", c):
                # Found a non-whitespace character before `(`.
                # Assume this is just a declaration line then and
                # skip to the closing bracket.
                _skip_to(f, ")")
                return None, None

    if key == "mixed":
        # From
        # <https://www.afs.enea.it/project/neptunius/docs/fluent/html/ug/node1470.htm>:
        #
        # > If a zone is of mixed type (element-type=0), it will have a body that
        # > lists the element type of each cell.
        #
        # No idea where the information other than the element types is stored
        # though. Skip for now.
        data = None
    else:
        # read cell data
        if out.group(1) == "":
            # ASCII cells
            data = np.empty((num_cells, num_nodes_per_cell), dtype=int)
            for k in range(num_cells):
                line = f.readline().decode()
                dat = line.split()
                if len(dat) != num_nodes_per_cell:
                    raise ReadError()
                data[k] = [int(d, 16) for d in dat]
        else:
            if key == "mixed":
                raise ReadError("Cannot read mixed cells in binary mode yet")
            # binary cells
            if out.group(1) == "20":
                dtype = np.int32
            else:
                if out.group(1) != "30":
                    ReadError(f"Expected keys '20' or '30', got {out.group(1)}.")
                dtype = np.int64
            shape = (num_cells, num_nodes_per_cell)
            count = shape[0] * shape[1]
            data = np.fromfile(f, count=count, dtype=dtype).reshape(shape)

    # make sure that the data set is properly closed
    _skip_close(f, 2)
    return key, data


def _read_faces(f, line):
    # faces
    # (13 (zone-id first-index last-index type element-type))

    # If the line is self-contained, it is merely a declaration of
    # the total number of points.
    if line.count("(") == line.count(")"):
        return {}

    out = re.match("\\s*\\(\\s*(|20|30)13\\s*\\(([^\\)]+)\\).*", line)
    assert out is not None
    a = [int(num, 16) for num in out.group(2).split()]

    if len(a) <= 4:
        raise ReadError()
    first_index = a[1]
    last_index = a[2]
    num_cells = last_index - first_index + 1
    element_type = a[4]

    element_type_to_key_num_nodes = {
        0: ("mixed", None),
        2: ("line", 2),
        3: ("triangle", 3),
        4: ("quad", 4),
    }

    key, num_nodes_per_cell = element_type_to_key_num_nodes[element_type]

    # Skip ahead to the line that opens the data block (might be
    # the current line already).
    if line.strip()[-1] != "(":
        _skip_to(f, "(")

    data = {}
    if out.group(1) == "":
        # ASCII
        if key == "mixed":
            # From
            # <https://www.afs.enea.it/project/neptunius/docs/fluent/html/ug/node1471.htm>:
            #
            # > If the face zone is of mixed type (element-type = > 0), the body of the
            # > section will include the face type and will appear as follows
            # >
            # > type v0 v1 v2 c0 c1
            # >
            for k in range(num_cells):
                line = ""
                while line.strip() == "":
                    line = f.readline().decode()
                dat = line.split()
                type_index = int(dat[0], 16)
                if type_index == 0:
                    raise ReadError()
                type_string, num_nodes_per_cell = element_type_to_key_num_nodes[
                    type_index
                ]
                if len(dat) != num_nodes_per_cell + 3:
                    raise ReadError()

                if type_string not in data:
                    data[type_string] = []

                data[type_string].append(
                    [int(d, 16) for d in dat[1 : num_nodes_per_cell + 1]]
                )

            data = {key: np.array(data[key]) for key in data}

        else:
            # read cell data
            data = np.empty((num_cells, num_nodes_per_cell), dtype=int)
            for k in range(num_cells):
                line = f.readline().decode()
                dat = line.split()
                # The body of a regular face section contains the grid connectivity, and
                # each line appears as follows:
                #   n0 n1 n2 cr cl
                # where n* are the defining nodes (vertices) of the face, and c* are the
                # adjacent cells.
                if len(dat) != num_nodes_per_cell + 2:
                    raise ReadError()
                data[k] = [int(d, 16) for d in dat[:num_nodes_per_cell]]
            data = {key: data}
    else:
        # binary
        if out.group(1) == "20":
            dtype = np.int32
        else:
            if out.group(1) != "30":
                ReadError(f"Expected keys '20' or '30', got {out.group(1)}.")
            dtype = np.int64

        if key == "mixed":
            raise ReadError("Mixed element type for binary faces not supported yet")

        # Read cell data.
        # The body of a regular face section contains the grid
        # connectivity, and each line appears as follows:
        #   n0 n1 n2 cr cl
        # where n* are the defining nodes (vertices) of the face,
        # and c* are the adjacent cells.
        shape = (num_cells, num_nodes_per_cell + 2)
        count = shape[0] * shape[1]
        data = np.fromfile(f, count=count, dtype=dtype).reshape(shape)
        # Cut off the adjacent cell data.
        data = data[:, :num_nodes_per_cell]
        data = {key: data}

    # make sure that the data set is properly closed
    _skip_close(f, 2)

    return data


def read(filename):  # noqa: C901
    # Initialize the data optional data fields
    field_data = {}
    cell_data = {}
    point_data = {}

    points = []
    cells = []

    first_point_index_overall = None
    last_point_index = None

    # read file in binary mode since some data might be binary
    with open_file(filename, "rb") as f:
        while True:
            line = f.readline().decode()
            if not line:
                break

            if line.strip() == "":
                continue

            # expect the line to have the form
            #  (<index> [...]
            out = re.match("\\s*\\(\\s*([0-9]+).*", line)
            if not out:
                raise ReadError()
            index = out.group(1)

            if index == "0":
                # Comment.
                _skip_close(f, line.count("(") - line.count(")"))
            elif index == "1":
                # header
                # (1 "<text>")
                _skip_close(f, line.count("(") - line.count(")"))
            elif index == "2":
                # dimensionality
                # (2 3)
                _skip_close(f, line.count("(") - line.count(")"))
            elif re.match("(|20|30)10", index):
                # points
                pts, first_point_index_overall, last_point_index = _read_points(
                    f, line, first_point_index_overall, last_point_index
                )

                if pts is not None:
                    points.append(pts)

            elif re.match("(|20|30)12", index):
                # cells
                # (2012 (zone-id first-index last-index type element-type))
                key, data = _read_cells(f, line)
                if data is not None:
                    cells.append((key, data))

            elif re.match("(|20|30)13", index):
                data = _read_faces(f, line)

                for key in data:
                    cells.append((key, data[key]))

            elif index == "39":
                warn("Zone specification not supported yet. Skipping.")
                _skip_close(f, line.count("(") - line.count(")"))

            elif index == "45":
                # (45 (2 fluid solid)())
                obj = re.match("\\(45 \\([0-9]+ ([\\S]+) ([\\S]+)\\)\\(\\)\\)", line)
                if obj:
                    warn(
                        f"Zone specification not supported yet ({obj.group(1)}, {obj.group(2)}). "
                        + "Skipping.",
                    )
                else:
                    warn("Zone specification not supported yet.")

            else:
                warn(f"Unknown index {index}. Skipping.")
                # Skipping ahead to the next line with two closing brackets.
                _skip_close(f, line.count("(") - line.count(")"))

    points = np.concatenate(points)

    # Gauge the cells with the first point_index.
    for k, c in enumerate(cells):
        cells[k] = (c[0], c[1] - first_point_index_overall)

    return Mesh(
        points, cells, point_data=point_data, cell_data=cell_data, field_data=field_data
    )


def write(filename, mesh, binary=True):
    with open_file(filename, "wb") as fh:
        # header
        fh.write(f'(1 "meshio {__version__}")\n'.encode())

        # dimension
        num_points, dim = mesh.points.shape
        if dim not in [2, 3]:
            raise WriteError(f"Can only write dimension 2, 3, got {dim}.")
        fh.write((f"(2 {dim})\n").encode())

        # total number of nodes
        first_node_index = 1
        fh.write((f"(10 (0 {first_node_index:x} {num_points:x} 0))\n").encode())

        # total number of cells
        total_num_cells = sum(len(c) for c in mesh.cells)
        fh.write((f"(12 (0 1 {total_num_cells:x} 0))\n").encode())

        # Write nodes
        key = "3010" if binary else "10"
        fh.write(
            f"({key} (1 {first_node_index:x} {num_points:x} 1 {dim:x})(\n".encode()
        )
        if binary:
            mesh.points.tofile(fh)
            fh.write(b"\n)")
            fh.write(b"End of Binary Section 3010)\n")
        else:
            np.savetxt(fh, mesh.points, fmt="%.16e")
            fh.write(b"))\n")

        # Write cells
        meshio_to_ansys_type = {
            # "mixed": 0,
            "triangle": 1,
            "tetra": 2,
            "quad": 3,
            "hexahedron": 4,
            "pyramid": 5,
            "wedge": 6,
            # "polyhedral": 7,
        }
        first_index = 0
        binary_dtypes = {
            # np.int16 is not allowed
            np.dtype("int32"): "2012",
            np.dtype("int64"): "3012",
        }
        for cell_block in mesh.cells:
            cell_type = cell_block.type
            values = cell_block.data
            key = binary_dtypes[values.dtype] if binary else "12"
            last_index = first_index + len(values) - 1
            try:
                ansys_cell_type = meshio_to_ansys_type[cell_type]
            except KeyError:
                legal_keys = ", ".join(meshio_to_ansys_type.keys())
                raise KeyError(
                    f"Illegal ANSYS cell type '{cell_type}'. (legal: {legal_keys})"
                )
            fh.write(
                f"({key} (1 {first_index:x} {last_index:x} 1 {ansys_cell_type})(\n".encode()
            )
            if binary:
                (values + first_node_index).tofile(fh)
                fh.write(b"\n)")
                fh.write((f"End of Binary Section {key})\n").encode())
            else:
                np.savetxt(fh, values + first_node_index, fmt="%x")
                fh.write(b"))\n")
            first_index = last_index + 1


register_format("ansys", [".msh"], read, {"ansys": write})



"""
I/O for DOLFIN's XML format, cf.
<https://people.sc.fsu.edu/~jburkardt/data/dolfin_xml/dolfin_xml.html>.
"""
import os
import pathlib
import re
from xml.etree import ElementTree as ET

import numpy as np

from .._common import warn
from .._exceptions import ReadError, WriteError
from .._helpers import register_format
from .._mesh import Mesh


def _read_mesh(filename):
    dolfin_to_meshio_type = {"triangle": ("triangle", 3), "tetrahedron": ("tetra", 4)}

    # Use iterparse() to avoid loading the entire file via parse(). iterparse()
    # allows to discard elements (via clear()) after they have been processed.
    # See <https://stackoverflow.com/a/326541/353337>.
    dim = None
    points = None
    keys = None
    cell_type = None
    num_nodes_per_cell = None
    cells = None
    cell_tags = None
    for event, elem in ET.iterparse(filename, events=("start", "end")):
        if event == "end":
            continue

        if elem.tag == "dolfin":
            # Don't be too strict with the assertion. Some mesh files don't have the
            # proper tags.
            # assert elem.attrib['nsmap'] \
            #     == '{\'dolfin\': \'https://fenicsproject.org/\'}'
            pass
        elif elem.tag == "mesh":
            dim = int(elem.attrib["dim"])
            cell_type, num_nodes_per_cell = dolfin_to_meshio_type[
                elem.attrib["celltype"]
            ]
            cell_tags = [f"v{i}" for i in range(num_nodes_per_cell)]
        elif elem.tag == "vertices":
            if dim is None:
                raise ReadError("Expected `mesh` before `vertices`")
            points = np.empty((int(elem.attrib["size"]), dim))
            keys = ["x", "y"]
            if dim == 3:
                keys += ["z"]
        elif elem.tag == "vertex":
            if points is None or keys is None:
                raise ReadError("Expected `vertices` before `vertex`")
            k = int(elem.attrib["index"])
            points[k] = [elem.attrib[key] for key in keys]
        elif elem.tag == "cells":
            if cell_type is None or num_nodes_per_cell is None:
                raise ReadError("Expected `mesh` before `cells`")
            cells = [
                (
                    cell_type,
                    np.empty((int(elem.attrib["size"]), num_nodes_per_cell), dtype=int),
                )
            ]
        elif elem.tag in ["triangle", "tetrahedron"]:
            k = int(elem.attrib["index"])
            assert cells is not None
            assert cell_tags is not None
            cells[0][1][k] = [elem.attrib[t] for t in cell_tags]
        else:
            warn(f"Unknown entry {elem.tag}. Ignoring.")

        elem.clear()

    return points, cells, cell_type


def _read_cell_data(filename):
    dolfin_type_to_numpy_type = {
        "int": np.dtype("int"),
        "float": np.dtype("float"),
        "uint": np.dtype("uint"),
    }

    cell_data = {}
    dir_name = pathlib.Path(filename).resolve().parent

    # Loop over all files in the same directory as `filename`.
    basename = pathlib.Path(filename).stem
    for f in os.listdir(dir_name):
        # Check if there are files by the name "<filename>_*.xml"; if yes,
        # extract the * pattern and make it the name of the data set.
        out = re.match(f"{basename}_([^\\.]+)\\.xml", f)
        if not out:
            continue
        name = out.group(1)

        parser = ET.XMLParser()
        tree = ET.parse((dir_name / f).as_posix(), parser)
        root = tree.getroot()

        mesh_functions = list(root)
        if len(mesh_functions) != 1:
            raise ReadError("Can only handle one mesh function")
        mesh_function = mesh_functions[0]

        if mesh_function.tag != "mesh_function":
            raise ReadError()
        size = int(mesh_function.attrib["size"])
        dtype = dolfin_type_to_numpy_type[mesh_function.attrib["type"]]
        data = np.empty(size, dtype=dtype)
        for child in mesh_function:
            if child.tag != "entity":
                raise ReadError()
            idx = int(child.attrib["index"])
            data[idx] = child.attrib["value"]

        if name not in cell_data:
            cell_data[name] = []
        cell_data[name].append(data)

    return cell_data


def read(filename):
    points, cells, _ = _read_mesh(filename)
    cell_data = _read_cell_data(filename)
    return Mesh(points, cells, cell_data=cell_data)


def _write_mesh(filename, points, cell_type, cells):
    stripped_cells = [c for c in cells if c.type == cell_type]

    meshio_to_dolfin_type = {"triangle": "triangle", "tetra": "tetrahedron"}

    if any(c.type != cell_type for c in cells):
        discarded_cell_types = {c.type for c in cells if c.type != cell_type}
        warn(
            "DOLFIN XML can only handle one cell type at a time. "
            + f"Using {cell_type}, discarding {', '.join(discarded_cell_types)}.",
        )

    dim = points.shape[1]
    if dim not in [2, 3]:
        raise WriteError(f"Can only write dimension 2, 3, got {dim}.")

    coord_names = ["x", "y"]
    if dim == 3:
        coord_names += ["z"]

    with open(filename, "w") as f:
        f.write("<dolfin nsmap=\"{'dolfin': 'https://fenicsproject.org/'}\">\n")
        ct = meshio_to_dolfin_type[cell_type]
        f.write(f'  <mesh celltype="{ct}" dim="{dim}">\n')

        num_points = len(points)
        f.write(f'    <vertices size="{num_points}">\n')
        for idx, point in enumerate(points):
            s = " ".join(f'{xyz}="{p}"' for xyz, p in zip("xyz", point))
            f.write(f'      <vertex index="{idx}" {s} />\n')
        f.write("    </vertices>\n")

        num_cells = 0
        for c in stripped_cells:
            num_cells += len(c.data)

        f.write(f'    <cells size="{num_cells}">\n')
        idx = 0
        for cell_block in stripped_cells:
            type_string = meshio_to_dolfin_type[cell_block.type]
            for cell in cell_block.data:
                s = " ".join(f'v{k}="{c}"' for k, c in enumerate(cell))
                f.write(f'      <{type_string} index="{idx}" {s} />\n')
                idx += 1
        f.write("    </cells>\n")
        f.write("  </mesh>\n")
        f.write("</dolfin>")


def _numpy_type_to_dolfin_type(dtype):
    types = {
        "int": [np.int8, np.int16, np.int32, np.int64],
        "uint": [np.uint8, np.uint16, np.uint32, np.uint64],
        "float": [np.float16, np.float32, np.float64],
    }
    for key, numpy_types in types.items():
        for numpy_type in numpy_types:
            if np.issubdtype(dtype, numpy_type):
                return key

    raise WriteError("Could not convert NumPy data type to DOLFIN data type.")


def _write_cell_data(filename, dim, cell_data):
    dolfin = ET.Element("dolfin", nsmap={"dolfin": "https://fenicsproject.org/"})

    mesh_function = ET.SubElement(
        dolfin,
        "mesh_function",
        type=_numpy_type_to_dolfin_type(cell_data.dtype),
        dim=str(dim),
        size=str(len(cell_data)),
    )

    for k, value in enumerate(cell_data):
        ET.SubElement(mesh_function, "entity", index=str(k), value=repr(value))

    tree = ET.ElementTree(dolfin)
    tree.write(filename)


def write(filename, mesh):
    warn("DOLFIN XML is a legacy format. Consider using XDMF instead.")

    if any("tetra" == c.type for c in mesh.cells):
        cell_type = "tetra"
    elif any("triangle" == c.type for c in mesh.cells):
        cell_type = "triangle"
    else:
        raise WriteError(
            "DOLFIN XML only supports triangles and tetrahedra. "
            "Consider using XDMF instead."
        )

    _write_mesh(filename, mesh.points, cell_type, mesh.cells)

    for name, lst in mesh.cell_data.items():
        for data in lst:
            fname = os.path.splitext(filename)[0]
            cell_data_filename = f"{fname}_{name}.xml"
            dim = 2 if mesh.points.shape[1] == 2 or all(mesh.points[:, 2] == 0) else 3
            _write_cell_data(cell_data_filename, dim, np.array(data))


register_format("dolfin-xml", [".xml"], read, {"dolfin-xml": write})



import numpy as np
from pathlib import Path
from typing import Union


extension = {
    ".flma": "FLMA",
    ".su2": "SU2",
    ".msh": "GMSH",
    ".vtk": "VTK",
    ".cgns": "CGNS",
    ".inp": "ABAQUS",
    ".obj": "OBJ",
}


class ControlDict:
    """
    input:\n
        Airfoil_file: a airfoil file contains coordinates
        Airfoil_data: a ndarray contains coordinates
        note: Airdoil_file and Airfoil_data only one can be defined

        Output: the output mesh name with suffix ( \n
            AVL FIRE : .flma,\n
            SU2 : .su2,\n
            GMSH : .msh,\n
            VTK : .vtk,\n
            CGNS : .cgns,\n
            ABAQUS : .inp,\n
            OBJ : .obj\n
            )\n
        
        "Airfoil_file":"./optimized_airfoil.dat",\n
        "Airfoil_data":null,\n
        "Output": "./Mesh/mesh.su2",\n
        "Spline_refine": {\n
            "Refinement tolerance": 172.0,\n
            "Refine trailing edge old": 3,\n
            "Refine trailing edge new": 6,\n
            "Refine trailing edge ratio": 3,\n
            "Number of points on spline": 200\n
        },\n
        "Trailing_edges":false,\n
        "Trailing_control": {\n
            "Upper side blending length": 30.0,\n
            "Lower side blending length": 30.0,\n
            "Upper blending polynomial exponent": 3,\n
            "Lower blending polynomial exponent": 3,\n
            "Trailing edge thickness relative to chord": 0.4\n
        },\n
        "Airfoil_mesh": {\n
            "Divisions normal to airfoil": 15,\n
            "1st cell layer thickness": 0.004,\n
            "Cell growth rate": 1.05\n
        },\n
        "Trailing_mesh": {\n
            "Divisions at trailing edge": 3,\n
            "Divisions downstream": 15,\n
            "1st cell layer thickness": 0.004,\n
            "Cell growth rate": 1.05\n
        },\n
        "Windtunnel_airfoil": {\n
            "Windtunnel height": 3.5,\n
            "Divisions of tunnel height": 100,\n
            "Cell thickness ratio": 10.0,\n
            "Distribution biasing": "symmetric",\n
            "Smoothing algorithm": "elliptic",\n
            "Smoothing iterations": 20,\n
            "Smoothing tolerance": 1.2e-5\n
        },\n
        "Windtunnel_wake": {\n
            "Windtunnel wake": 7.0,\n
            "Divisions in the wake": 100,\n
            "Cell thickness ratio": 15.0,\n
            "Equalize vertical wake line at": 30.0\n
        }\n

        other mesh control parameters can be find: \n
        https://pyaero.readthedocs.io/en/latest/

    """

    Airfoil_file: Union[str, Path] = None
    Airfoil_data: np.ndarray = None

    Output: str = None

    Spline_refine = {
        "Refinement tolerance": 172.0,
        "Refine trailing edge old": 3,
        "Refine trailing edge new": 6,
        "Refine trailing edge ratio": 3,
        "Number of points on spline": 200,
    }

    Trailing_edges: bool = False
    Trailing_control = {
        "Upper side blending length": 30.0,
        "Lower side blending length": 30.0,
        "Upper blending polynomial exponent": 3,
        "Lower blending polynomial exponent": 3,
        "Trailing edge thickness relative to chord": 0.4,
    }

    Airfoil_mesh = {
        "Divisions normal to airfoil": 15,
        "1st cell layer thickness": 0.004,
        "Cell growth rate": 1.05,
    }

    Trailing_mesh = {
        "Divisions at trailing edge": 3,
        "Divisions downstream": 15,
        "1st cell layer thickness": 0.004,
        "Cell growth rate": 1.05,
    }

    Windtunnel_airfoil = {
        "Windtunnel height": 3.5,
        "Divisions of tunnel height": 100,
        "Cell thickness ratio": 10.0,
        "Distribution biasing": "symmetric",
        "Smoothing algorithm": "elliptic",
        "Smoothing iterations": 20,
        "Smoothing tolerance": 1.2e-5,
    }

    Windtunnel_wake = {
        "Windtunnel wake": 7.0,
        "Divisions in the wake": 100,
        "Cell thickness ratio": 15.0,
        "Equalize vertical wake line at": 30.0,
    }



"""
I/O for Netgen mesh files
<https://github.com/NGSolve/netgen/blob/master/libsrc/meshing/meshclass.cpp>.
"""
import numpy as np

from ..__about__ import __version__
from .._common import warn
from .._files import open_file
from .._helpers import register_format
from .._mesh import Mesh


def _fast_forward_over_blank_lines(f):
    is_eof = False
    while True:
        line = f.readline()
        if not line:
            is_eof = True
            break

        line = line.strip()
        if len(line) > 0 and not line.startswith("#"):
            break
    return line, is_eof


netgen_codims = {"materials": 0, "bcnames": 1, "cd2names": 2, "cd3names": 3}


netgen0d_to_meshio_type = {
    1: "vertex",
}
netgen1d_to_meshio_type = {
    2: "line",
}
netgen2d_to_meshio_type = {
    3: "triangle",
    6: "triangle6",
    4: "quad",
    8: "quad8",
}

netgen3d_to_meshio_type = {
    4: "tetra",
    5: "pyramid",
    6: "wedge",
    8: "hexahedron",
    10: "tetra10",
    13: "pyramid13",
    15: "wedge15",
    20: "hexahedron20",
}

netgen_to_meshio_type = {
    0: netgen0d_to_meshio_type,
    1: netgen1d_to_meshio_type,
    2: netgen2d_to_meshio_type,
    3: netgen3d_to_meshio_type,
}

netgen_to_meshio_pmap = {
    "vertex": [0],
    "line": [0, 1],
    "triangle": list(range(3)),
    "triangle6": [0, 1, 2, 5, 3, 4],
    "quad": list(range(4)),
    "quad8": [0, 1, 2, 3, 4, 7, 5, 6],
    "tetra": [0, 2, 1, 3],
    "tetra10": [0, 2, 1, 3, 5, 7, 4, 6, 9, 8],
    "pyramid": [0, 3, 2, 1, 4],
    "pyramid13": [0, 3, 2, 1, 4, 7, 6, 8, 5, 9, 12, 11, 10],
    "wedge": [0, 2, 1, 3, 5, 4],
    "wedge15": [0, 2, 1, 3, 5, 4, 7, 8, 6, 13, 14, 12, 9, 11, 10],
    "hexahedron": [0, 3, 2, 1, 4, 7, 6, 5],
    "hexahedron20": [
        0,
        3,
        2,
        1,
        4,
        7,
        6,
        5,
        10,
        9,
        11,
        8,
        16,
        19,
        18,
        17,
        14,
        13,
        15,
        12,
    ],
}

meshio_to_netgen_pmap = {}
for t, pmap in netgen_to_meshio_pmap.items():
    n = len(pmap)
    ipmap = list(range(n))
    for i in range(n):
        ipmap[pmap[i]] = i
    meshio_to_netgen_pmap[t] = ipmap


def read(filename):
    if str(filename).endswith(".vol.gz"):
        import gzip

        with gzip.open(filename, "rt") as f:
            return read_buffer(f)

    with open_file(filename, "r") as f:
        return read_buffer(f)


def _read_cells(f, netgen_cell_type, cells, cells_index, skip_every_other_line=False):
    if netgen_cell_type == "pointelements":
        dim = 0
        nump = 1
        pi0 = 0
        i_index = 1
    elif netgen_cell_type.startswith("edgesegments"):
        dim = 1
        nump = 2
        pi0 = 2
        i_index = 0
    elif netgen_cell_type.startswith("surfaceelements"):
        dim = 2
        pi0 = 5
        i_index = 1
    elif netgen_cell_type == "volumeelements":
        dim = 3
        pi0 = 2
        i_index = 0
    else:
        raise ValueError(f"Unknown Netgen cell section '{netgen_cell_type}'")

    num_cells = int(f.readline())
    tmap = netgen_to_meshio_type[dim]

    for _ in range(num_cells):
        line, _ = _fast_forward_over_blank_lines(f)
        data = list(filter(None, line.split(" ")))
        index = int(data[i_index])
        if dim == 2:
            nump = int(data[4])
        elif dim == 3:
            nump = int(data[1])

        pi = list(map(int, data[pi0 : pi0 + nump]))

        t = tmap[nump]
        if len(cells) == 0 or t != cells[-1][0]:
            cells.append((t, []))
            cells_index.append([])
        cells[-1][1].append(pi)
        cells_index[-1].append(index)
        if skip_every_other_line:
            line, _ = _fast_forward_over_blank_lines(f)


def _write_cells(f, cell_block, index=None):
    if len(cell_block) == 0:
        return
    pmap = np.array(meshio_to_netgen_pmap[cell_block.type])
    post_data = []
    pre_data = []
    i_index = 0
    if cell_block.dim == 0:
        post_data = [1]
        i_index = 1
    elif cell_block.dim == 1:
        pre_data = [1, 0]
        post_data = [-1, -1, 0, 0, 1, 0, 1, 0]
    elif cell_block.dim == 2:
        pre_data = [1, 1, 0, 0, len(pmap)]
        i_index = 1
    elif cell_block.dim == 3:
        pre_data = [1, len(pmap)]
    else:
        raise ValueError(f"Invalid cell dimension: {cell_block.dim}")

    col1 = len(pre_data)
    col2 = col1 + len(pmap)
    col3 = col2 + len(post_data)

    pi = np.zeros((len(cell_block), col3), dtype=np.int32)
    pi[:, :col1] = np.repeat([pre_data], len(cell_block), axis=0)
    pi[:, col1:col2] = cell_block.data[:, pmap] + 1
    pi[:, col2:] = np.repeat([post_data], len(cell_block), axis=0)
    if index is not None:
        pi[:, i_index] = index
    np.savetxt(f, pi, "%i")


def _skip_block(f):
    n = int(f.readline())
    for _ in range(n):
        f.readline()


def _write_codim_domain_data(f, mesh, cells_index, dim, codim):
    # assume format as read from gmsh 4.1 files
    data = {}
    for name, val in mesh.field_data.items():
        if val[1] == dim - codim:
            data[val[0]] = name

    # set generic default names (is this appropriate/useful?)
    if len(data) == 0:
        indices = set()
        for cell_block, index in zip(mesh.cells, cells_index):
            if index is None:
                continue
            if cell_block.dim == dim - codim:
                indices = indices.union(set(index))

        for idx in indices:
            data[idx] = f"cd{codim:d}_{idx:d}"

    if len(data) == 0:
        return

    codim_tag = [kk for kk, vv in netgen_codims.items() if vv == codim][0]
    f.write(f"\n{codim_tag:s}\n")

    ncd = max(data.keys())
    f.write(f"{ncd:d}\n")
    for idx in range(1, ncd + 1):
        f.write("{:d} {:s}\n".format(idx, data.get(idx, "")))


def read_buffer(f):
    points = []
    cells = []
    cells_index = []
    field_data = {}
    identifications = None
    identificationtypes = None

    have_edgesegmentsgi2_in_two_lines = False

    line, is_eof = _fast_forward_over_blank_lines(f)
    if line != "mesh3d":
        raise RuntimeError("Not a valid Netgen mesh")

    while True:
        line, is_eof = _fast_forward_over_blank_lines(f)
        if is_eof:
            break

        elif line == "dimension":
            dimension = int(f.readline())
        elif line == "geomtype":
            geomtype = int(f.readline())
            if geomtype not in [0, 1, 10, 11, 12, 13]:
                warn(f"Unknown geomtype in Netgen mesh: {geomtype}")

        elif line == "points":
            num_points = int(f.readline())
            if num_points > 0:
                points = np.loadtxt(f, max_rows=num_points)
                if dimension != 3:
                    points = points[:, :dimension]

        elif line in [
            "pointelements",
            "edgesegments",
            "edgesegmentsgi",
            "edgesegmentsgi2",
            "surfaceelements",
            "surfaceelementsgi",
            "surfaceelementsuv",
            "volumeelements",
        ]:
            _read_cells(f, line, cells, cells_index, have_edgesegmentsgi2_in_two_lines)
        elif line == "endmesh":
            break

        elif line.split() == ["surf1", "surf2", "p1", "p2"]:
            # if this line is present, the edgesegmentsgi2 info is split in two lines per data set
            have_edgesegmentsgi2_in_two_lines = True

        elif line in netgen_codims.keys():
            edim = dimension - netgen_codims[line]
            num_entries = int(f.readline())
            for _ in range(num_entries):
                line = f.readline().split()
                if len(line) != 2:
                    continue
                idx, name = line
                field_data[name] = [int(idx), edim]

        elif line == "identifications":
            num_entries = int(f.readline())
            if num_entries > 0:
                identifications = np.loadtxt(
                    f, max_rows=num_entries, dtype=int
                ).reshape(num_entries, 3)

        elif line == "identificationtypes":
            num_entries = int(f.readline())
            if num_entries > 0:
                identificationtypes = np.loadtxt(f, max_rows=1, dtype=int).reshape(
                    1, num_entries
                )

        elif line in [
            "face_colours",
            "singular_edge_left",
            "singular_edge_right",
            "singular_face_inside",
            "singular_face_outside",
            "singular_points",
        ]:
            _skip_block(f)
        else:
            raise RuntimeError(f"Unknown Netgen mesh token: {line}")

    # convert to numpy arrays
    # subtract one (netgen counts 1-based)
    # apply permutation of vertex numbers
    for k, (t, data) in enumerate(cells):
        pmap = netgen_to_meshio_pmap[t]
        d = np.array(data, dtype=np.uint32)
        d[:, :] = d[:, pmap] - 1
        cells[k] = (t, d)

    # currently, there is no better place for identification data
    kwargs = {}
    if identifications is not None:
        kwargs["info"] = {
            "netgen:identifications": identifications,
            "netgen:identificationtypes": identificationtypes,
        }

    mesh = Mesh(
        points,
        cells,
        cell_data={"netgen:index": cells_index},
        field_data=field_data,
        **kwargs,
    )
    return mesh


def write(filename, mesh, float_fmt=".16e"):
    if str(filename).endswith(".vol.gz"):
        import gzip

        with gzip.open(filename, "wt") as f:
            write_buffer(f, mesh, float_fmt)
        return

    with open_file(filename, "w") as f:
        write_buffer(f, mesh, float_fmt)


def write_buffer(f, mesh, float_fmt):
    _, dimension = mesh.points.shape
    cells_per_dim = [0, 0, 0, 0]

    # Netgen can store one cell_index, i.e., integer cell data. Pick one in
    # mesh.cell_data, and prefer "netgen:index" if present. Unfortunately, netgen cannot
    # store the name of the data; when reading, it will always be "netgen:index".
    # See also <https://github.com/nschloe/meshio/issues/1199>.
    if "netgen:index" in mesh.cell_data:
        cells_index = mesh.cell_data["netgen:index"]
    else:
        # any other integer cell data?
        cells_index = None
        for values in mesh.cell_data.values():
            if np.issubdtype(values[0].dtype, np.integer):
                cells_index = values
                break

        if cells_index is None:
            cells_index = [None] * len(mesh.cells)

    for cell_block in mesh.cells:
        cells_per_dim[cell_block.dim] += len(cell_block)

    f.write(f"# Generated by meshio {__version__}\n")
    f.write("mesh3d\n\n")

    f.write("dimension\n")
    f.write(f"{dimension}\n\n")

    f.write("geomtype\n")
    f.write("0\n")

    f.write("\n# surfnr    bcnr   domin  domout      np      p1      p2      p3\n")
    f.write("surfaceelements\n")
    f.write(f"{cells_per_dim[2]}\n")
    for cell_block, index in zip(mesh.cells, cells_index):
        if cell_block.dim == 2:
            _write_cells(f, cell_block, index)

    f.write("\n#  matnr      np      p1      p2      p3      p4\n")
    f.write("volumeelements\n")
    f.write(f"{cells_per_dim[3]}\n")
    for cell_block, index in zip(mesh.cells, cells_index):
        if cell_block.dim == 3:
            _write_cells(f, cell_block, index)

    f.write(
        "\n# surfid  0   p1   p2   trignum1    trignum2   domin/surfnr1    domout/surfnr2   ednr1   dist1   ednr2   dist2\n",
    )
    f.write("edgesegmentsgi2\n")
    f.write(f"{cells_per_dim[1]}\n")
    for cell_block, index in zip(mesh.cells, cells_index):
        if cell_block.dim == 1:
            _write_cells(f, cell_block, index)

    f.write("\n#          X             Y             Z\n")
    f.write("points\n")
    f.write(f"{len(mesh.points)}\n")

    points = mesh.points
    if dimension != 3:
        points = np.hstack(
            (points, np.zeros((points.shape[0], 3 - dimension), dtype=points.dtype))
        )
    np.savetxt(f, points, "%" + float_fmt)

    f.write("\n#          pnum             index\n")
    f.write("pointelements\n")
    f.write(f"{cells_per_dim[0]}\n")
    for cell_block, index in zip(mesh.cells, cells_index):
        if cell_block.dim == 0:
            _write_cells(f, cell_block, index)

    # currently, there is no better place for identification data
    if isinstance(mesh.info, dict):
        identifications = mesh.info.get("netgen:identifications")
        identificationtypes = mesh.info.get("netgen:identificationtypes")
        if identifications is not None and identificationtypes is not None:
            f.write("\nidentifications\n")
            f.write(f"{identifications.shape[0]}\n")
            np.savetxt(f, identifications, "%d")
            f.write("\nidentificationtypes\n")
            f.write(f"{identificationtypes.size}\n")
            np.savetxt(
                f, identificationtypes.reshape(1, identificationtypes.size), "%d"
            )

    for codim in range(dimension + 1):
        _write_codim_domain_data(f, mesh, cells_index, dimension, codim)

    f.write("\nendmesh\n")


register_format("netgen", [".vol", ".vol.gz"], read, {"netgen": write})



from __future__ import annotations

from xml.etree import ElementTree as ET

import numpy as np

from .._exceptions import WriteError
from .._helpers import register_format


def write(
    filename,
    mesh,
    float_fmt: str = ".3f",
    stroke_width: str | None = None,
    # Use a default image_width (not None). If set to None, images will come out at the
    # width of the mesh (which is okay). Some viewers (e.g., eog) have problems
    # displaying SVGs of width around 1 since they interpret it as the width in pixels.
    image_width: int | float | None = 100,
    # ParaView's default colors
    fill: str = "#c8c5bd",
    stroke: str = "#000080",
):
    if mesh.points.shape[1] == 3 and not np.allclose(
        mesh.points[:, 2], 0.0, rtol=0.0, atol=1.0e-14
    ):
        raise WriteError(
            f"SVG can only handle flat 2D meshes (shape: {mesh.points.shape})"
        )

    pts = mesh.points[:, :2].copy()
    min_x = np.min(pts[:, 0]) if len(pts) > 0 else 0.0
    max_x = np.max(pts[:, 0]) if len(pts) > 0 else 0.0
    min_y = np.min(pts[:, 1]) if len(pts) > 0 else 0.0
    max_y = np.max(pts[:, 1]) if len(pts) > 0 else 0.0

    pts[:, 1] = max_y + min_y - pts[:, 1]

    width = max_x - min_x
    height = max_y - min_y

    if image_width is not None and width != 0:
        scaling_factor = image_width / width
        min_x *= scaling_factor
        min_y *= scaling_factor
        width *= scaling_factor
        height *= scaling_factor
        pts *= scaling_factor

    if stroke_width is None:
        stroke_width = str(width / 100)

    fmt = " ".join(4 * [f"{{:{float_fmt}}}"])
    svg = ET.Element(
        "svg",
        xmlns="http://www.w3.org/2000/svg",
        version="1.1",
        viewBox=fmt.format(min_x, min_y, width, height),
    )

    style = ET.SubElement(svg, "style")
    opts = [
        f"fill: {fill}",
        f"stroke: {stroke}",
        f"stroke-width: {stroke_width}",
        "stroke-linejoin:bevel",
    ]
    # Use path, not polygon, because svgo converts polygons to paths and doesn't convert
    # the style alongside. No problem if it's paths all the way.
    style.text = "path {" + "; ".join(opts) + "}"

    for cell_block in mesh.cells:
        if cell_block.type not in ["line", "triangle", "quad"]:
            continue

        if cell_block.type == "line":
            fmt = (
                f"M {{:{float_fmt}}} {{:{float_fmt}}}"
                + f"L {{:{float_fmt}}} {{:{float_fmt}}}"
            )
        elif cell_block.type == "triangle":
            fmt = (
                f"M {{:{float_fmt}}} {{:{float_fmt}}}"
                + f"L {{:{float_fmt}}} {{:{float_fmt}}}"
                + f"L {{:{float_fmt}}} {{:{float_fmt}}}"
                + "Z"
            )
        elif cell_block.type == "quad":
            fmt = (
                f"M {{:{float_fmt}}} {{:{float_fmt}}}"
                + f"L {{:{float_fmt}}} {{:{float_fmt}}}"
                + f"L {{:{float_fmt}}} {{:{float_fmt}}}"
                + f"L {{:{float_fmt}}} {{:{float_fmt}}}"
                + "Z"
            )
        for cell in cell_block.data:
            ET.SubElement(
                svg,
                "path",
                d=fmt.format(*pts[cell].flatten()),
            )

    tree = ET.ElementTree(svg)
    tree.write(filename)


register_format("svg", [".svg"], None, {"svg": write})



# This XML writer is a drop-in replacement for LXML/Python XML Etree. It only offers one
# other member: self.text_write.
# The problem is that, for LXML, the entire etree has to be constructed in memory before
# writing it to a file. Many mesh formats that use XML have lots of int or float data
# written in the text fields. Converting this to ASCII first requires a lot of memory.
# This etree here allows the writing method to write to the file directly, without
# having to create a string representation first.


class Element:
    def __init__(self, name, **kwargs):
        self.name = name
        self.attrib = kwargs
        self._children = []
        self.text = None
        self.text_writer = None

    def insert(self, pos, elem):
        self._children.insert(pos, elem)

    def set(self, key, value):
        self.attrib[key] = value

    def write(self, f):
        kw_list = [f'{key}="{value}"' for key, value in self.attrib.items()]
        f.write("<{}>\n".format(" ".join([self.name] + kw_list)))
        if self.text:
            f.write(self.text)
            f.write("\n")
        if self.text_writer:
            self.text_writer(f)
            f.write("\n")
        for child in self._children:
            child.write(f)
        f.write(f"</{self.name}>\n")


class SubElement(Element):
    def __init__(self, parent, name, **kwargs):
        super().__init__(name, **kwargs)
        parent._children.append(self)


class Comment:
    def __init__(self, text):
        self.text = text

    def write(self, f):
        f.write(f"<!--{self.text}-->\n")


class ElementTree:
    def __init__(self, root):
        self.root = root

    def write(self, filename, xml_declaration=True):
        with open(filename, "w") as f:
            if xml_declaration:
                f.write('<?xml version="1.0"?>\n')
            self.root.write(f)



"""
I/O for the STL format, cf.
<https://en.wikipedia.org/wiki/STL_(file_format)>.
"""
from __future__ import annotations

import os

import numpy as np

from ..__about__ import __version__
from .._common import warn
from .._exceptions import ReadError
from .._files import open_file
from .._helpers import register_format
from .._mesh import CellBlock, Mesh


def read(filename):
    with open_file(filename, "rb") as f:
        # Checking if the file is ASCII format is normally done by checking if the
        # first 5 characters of the header is "solid".
        # ```
        # header = f.read(80).decode()
        # ```
        # Unfortunately, there are mesh files out there which are binary and still put
        # "solid" there.
        # A suggested alternative is to pretend the file is binary, read the
        # num_triangles and see if it matches the file size
        # (https://stackoverflow.com/a/7394842/353337).
        filesize_bytes = os.path.getsize(filename)
        if filesize_bytes < 80:
            return _read_ascii(f)

        f.read(80)
        num_triangles = np.fromfile(f, count=1, dtype="<u4")[0]
        # for each triangle, one has 3 float32 (facet normal), 9 float32 (facet),
        # and 1 int16 (attribute count), 50 bytes in total
        if 84 + num_triangles * 50 == filesize_bytes:
            return _read_binary(f, num_triangles)

        # rewind and skip header
        f.seek(0)
        f.readline()
        return _read_ascii(f)


# np.loadtxt is super slow
# Code adapted from <https://stackoverflow.com/a/8964779/353337>.
def iter_loadtxt(
    infile,
    skiprows: int = 0,
    comments: str | tuple[str, ...] = "#",
    dtype=float,
    usecols: tuple[int] | None = None,
):
    def iter_func():
        items = None
        for _ in range(skiprows):
            try:
                next(infile)
            except StopIteration:
                raise ReadError("EOF Skipped too many rows")

        for line in infile:
            line = line.decode().strip()
            if line.startswith(comments):
                continue
            # remove all text
            items = line.split()[-3:]
            usecols_ = range(len(items)) if usecols is None else usecols
            for idx in usecols_:
                yield dtype(items[idx])

        if items is None:
            iter_loadtxt.rowlength = 3
            return

        iter_loadtxt.rowlength = len(items) if usecols is None else len(usecols)

    data = np.fromiter(iter_func(), dtype=dtype)
    return data.reshape((-1, iter_loadtxt.rowlength))


def _read_ascii(f):
    # The file has the form
    # ```
    # solid foo
    #   facet normal 0.455194 -0.187301 -0.870469
    #    outer loop
    #     vertex 266.36 234.594 14.6145
    #     vertex 268.582 234.968 15.6956
    #     vertex 267.689 232.646 15.7283
    #    endloop
    #   endfacet
    #   # [...] more facets [...]
    # endsolid
    # ```
    # In the interest of speed, don't verify the format and instead just skip the text.

    # TODO Pandas is MUCH faster than numpy for i/o, see
    # <https://stackoverflow.com/a/18260092/353337>.
    # import pandas
    # data = pandas.read_csv(
    #     f,
    #     skiprows=lambda row: row == 0 or (row - 1) % 7 in [0, 1, 5, 6],
    #     skipfooter=1,
    #     usecols=(1, 2, 3),
    # )

    # np.loadtxt is super slow
    # data = np.loadtxt(
    #     f,
    #     comments=["solid", "facet", "outer loop", "endloop", "endfacet", "endsolid"],
    #     usecols=(1, 2, 3),
    # )
    data = iter_loadtxt(
        f,
        comments=("solid", "outer loop", "endloop", "endfacet", "endsolid"),
        # usecols=(1, 2, 3),
    )

    if data.shape[0] % 4 != 0:
        raise ReadError()

    # split off the facet normals
    facet_rows = np.zeros(len(data), dtype=bool)
    facet_rows[0::4] = True
    facet_normals = data[facet_rows]
    data = data[~facet_rows]

    if data.shape[0] == 0:
        points = []
        cells = {}
        cell_data = {}
    else:
        facets = np.split(data, data.shape[0] // 3)
        points, cells = data_from_facets(facets)
        cell_data = {"facet_normals": [facet_normals]}

    return Mesh(points, cells, cell_data=cell_data)


def data_from_facets(facets):
    # Now, all facets contain the point coordinate. Try to identify individual points
    # and build the data arrays.
    if len(facets) == 0:
        points = np.empty((0, 3), dtype=float)
        cells = []
    else:
        pts = np.concatenate(facets)
        # TODO equip `unique()` with a tolerance
        # Use return_index so we can use sort on `idx` such that the order is
        # preserved; see <https://stackoverflow.com/a/15637512/353337>.
        _, idx, inv = np.unique(pts, axis=0, return_index=True, return_inverse=True)
        k = np.argsort(idx)
        points = pts[idx[k]]
        inv_k = np.argsort(k)
        cells = [CellBlock("triangle", inv_k[inv].reshape(-1, 3))]
    return points, cells


def _read_binary(f, num_triangles: int):
    # for each triangle, one has 3 float32 (facet normal), 9 float32 (facet), and 1
    # int16 (attribute count)
    out = np.fromfile(
        f,
        count=num_triangles,
        dtype=np.dtype(
            [("normal", "<f4", (3,)), ("facet", "<f4", (3, 3)), ("attr count", "<i2")]
        ),
    )
    # discard normals, attribute count
    facets = out["facet"]
    # if not np.all(out["attr count"] == 0):
    #     print(out["attr count"])
    #     raise ReadError("Nonzero attr count")

    points, cells = data_from_facets(facets)
    return Mesh(points, cells)


def write(filename, mesh, binary=False):
    if "triangle" not in {block.type for block in mesh.cells}:
        warn("STL can only write triangle cells. No triangle cells found.")
    if len(mesh.cells) > 1:
        invalid = {block.type for block in mesh.cells if block.type != "triangle"}
        invalid = ", ".join(invalid)
        warn(f"STL can only write triangle cells. Discarding {invalid}.")

    if mesh.points.shape[1] == 2:
        warn(
            "STL requires 3D points, but 2D points given. Appending 0 third component."
        )
        points = np.column_stack([mesh.points, np.zeros_like(mesh.points[:, 0])])
    else:
        points = mesh.points

    pts = points[mesh.get_cells_type("triangle")]
    if "facet_normals" in mesh.cell_data:
        normals = mesh.get_cell_data("facet_normals", "triangle")
    else:
        normals = np.cross(pts[:, 1] - pts[:, 0], pts[:, 2] - pts[:, 0])
        nrm = np.sqrt(np.einsum("ij,ij->i", normals, normals))
        normals = (normals.T / nrm).T

    fun = _write_binary if binary else _write_ascii
    fun(filename, pts, normals)


def _write_ascii(filename, pts, normals):
    with open_file(filename, "w") as fh:
        fh.write("solid\n")
        for local_pts, normal in zip(pts, normals):
            out = (
                "\n".join(
                    [
                        "facet normal {} {} {}".format(*normal),
                        " outer loop",
                        "  vertex {} {} {}".format(*local_pts[0]),
                        "  vertex {} {} {}".format(*local_pts[1]),
                        "  vertex {} {} {}".format(*local_pts[2]),
                        " endloop",
                        "endfacet",
                    ]
                )
                + "\n"
            )
            fh.write(out)
        fh.write("endsolid\n")


def _write_binary(filename, pts, normals):
    with open_file(filename, "wb") as fh:
        # 80 character header data
        msg = f"This file was generated by meshio v{__version__}."
        msg += (79 - len(msg)) * "X"
        msg += "\n"
        fh.write(msg.encode())

        fh.write(np.array(len(pts)).astype("<u4"))

        dtype = np.dtype(
            [
                ("normal", ("<f4", 3)),
                ("points", ("<f4", (3, 3))),
                ("attr", "<u2"),
            ]
        )
        a = np.empty(len(pts), dtype=dtype)
        a["normal"] = normals
        a["points"] = pts
        a["attr"] = 0
        a.tofile(fh)


register_format("stl", [".stl"], read, {"stl": write})



"""
I/O for AFLR's UGRID format
[1] <https://www.simcenter.msstate.edu/software/documentation/ug_io/3d_grid_file_type_ugrid.html>
Check out
[2] <https://www.simcenter.msstate.edu/software/public/ug_io/index_simsys_web.php>
for UG_IO C code able to read and convert UGRID files
Node ordering described in
[3] <https://www.simcenter.msstate.edu/software/documentation/ug_io/3d_input_output_grids.html>
"""
import numpy as np

from .._common import _pick_first_int_data, warn
from .._exceptions import ReadError
from .._files import open_file
from .._helpers import register_format
from .._mesh import CellBlock, Mesh

# Float size and endianness are recorded by these suffixes
# binary files come in C-type or FORTRAN type
# https://www.simcenter.msstate.edu/software/documentation/ug_io/ugc_file_formats.html
#
# 64-bit versions described here
# https://www.simcenter.msstate.edu/software/documentation/ug_io/ugc_l_file_formats.html
file_types = {
    "ascii": {"type": "ascii", "float_type": "f", "int_type": "i"},
    "b8l": {"type": "C", "float_type": ">f8", "int_type": ">i8"},
    "b8": {"type": "C", "float_type": ">f8", "int_type": ">i4"},
    "b4": {"type": "C", "float_type": ">f4", "int_type": ">i4"},
    "lb8l": {"type": "C", "float_type": "<f8", "int_type": "<i8"},
    "lb8": {"type": "C", "float_type": "<f8", "int_type": "<i4"},
    "lb4": {"type": "C", "float_type": "<f4", "int_type": "<i4"},
    "r8": {"type": "F", "float_type": ">f8", "int_type": ">i4"},
    "r4": {"type": "F", "float_type": ">f4", "int_type": ">i4"},
    "lr8": {"type": "F", "float_type": "<f8", "int_type": "<i4"},
    "lr4": {"type": "F", "float_type": "<f4", "int_type": "<i4"},
}


def determine_file_type(filename):
    file_type = file_types["ascii"]
    filename_parts = str(filename).split(".")
    if len(filename_parts) > 1:
        type_suffix = filename_parts[-2]
        if type_suffix in file_types.keys():
            file_type = file_types[type_suffix]
    return file_type


def read(filename):
    file_type = determine_file_type(filename)
    with open_file(filename) as f:
        mesh = read_buffer(f, file_type)
    return mesh


def _read_section(f, file_type, count, dtype):
    if file_type["type"] == "ascii":
        return np.fromfile(f, count=count, dtype=dtype, sep=" ")
    return np.fromfile(f, count=count, dtype=dtype)


def read_buffer(f, file_type):
    cells = []
    cell_data = []

    itype = file_type["int_type"]
    ftype = file_type["float_type"]

    # Fortran type includes a number of bytes before and after each record, according to
    # documentation [1] there are two records in the file; see also UG_IO freely
    # available code at [2].
    if file_type["type"] == "F":
        _read_section(f, file_type, count=1, dtype=itype)

    nitems = _read_section(f, file_type, count=7, dtype=itype)

    if file_type["type"] == "F":
        _read_section(f, file_type, count=1, dtype=itype)

    if not nitems.size == 7:
        raise ReadError("Header of ugrid file is ill-formed")

    ugrid_counts = {
        "points": (nitems[0], 3),
        "triangle": (nitems[1], 3),
        "quad": (nitems[2], 4),
        "tetra": (nitems[3], 4),
        "pyramid": (nitems[4], 5),
        "wedge": (nitems[5], 6),
        "hexahedron": (nitems[6], 8),
    }

    if file_type["type"] == "F":
        _read_section(f, file_type, count=1, dtype=itype)

    nnodes = ugrid_counts["points"][0]
    points = _read_section(f, file_type, count=nnodes * 3, dtype=ftype).reshape(
        nnodes, 3
    )

    for key in ["triangle", "quad"]:
        nitems = ugrid_counts[key][0]
        nvertices = ugrid_counts[key][1]
        if nitems == 0:
            continue
        out = _read_section(
            f, file_type, count=nitems * nvertices, dtype=itype
        ).reshape(nitems, nvertices)
        # UGRID is one-based
        cells.append(CellBlock(key, out - 1))

    cell_data = {"ugrid:ref": []}
    for key in ["triangle", "quad"]:
        nitems = ugrid_counts[key][0]
        if nitems == 0:
            continue
        out = _read_section(f, file_type, count=nitems, dtype=itype)
        cell_data["ugrid:ref"].append(out)

    for key in ["tetra", "pyramid", "wedge", "hexahedron"]:
        nitems = ugrid_counts[key][0]
        nvertices = ugrid_counts[key][1]
        if nitems == 0:
            continue
        out = _read_section(
            f, file_type, count=nitems * nvertices, dtype=itype
        ).reshape(nitems, nvertices)

        if key == "pyramid":
            out = out[:, [1, 0, 3, 4, 2]]

        # UGRID is one-based
        cells.append(CellBlock(key, out - 1))

        # fill volume element attributes with zero
        cell_data["ugrid:ref"].append(np.zeros(nitems, dtype=int))

    if file_type["type"] == "F":
        _read_section(f, file_type, count=1, dtype=itype)

    return Mesh(points, cells, cell_data=cell_data)


def _write_section(f, file_type, array, dtype):
    if file_type["type"] == "ascii":
        ncols = array.shape[1]
        fmt = " ".join(["%r"] * ncols)
        np.savetxt(f, array, fmt=fmt)
    else:
        array.astype(dtype).tofile(f)


def write(filename, mesh):
    file_type = determine_file_type(filename)

    with open_file(filename, "w") as f:
        _write_buffer(f, file_type, mesh)


def _write_buffer(f, file_type, mesh):
    itype = file_type["int_type"]
    ftype = file_type["float_type"]

    ugrid_counts = {
        "points": 0,
        "triangle": 0,
        "quad": 0,
        "tetra": 0,
        "pyramid": 0,
        "wedge": 0,
        "hexahedron": 0,
    }
    # ugrid type to cell array id
    ugrid_meshio_id = {
        "points": -1,
        "triangle": -1,
        "quad": -1,
        "tetra": -1,
        "pyramid": -1,
        "wedge": -1,
        "hexahedron": -1,
    }

    ugrid_counts["points"] = mesh.points.shape[0]

    for i, cell_block in enumerate(mesh.cells):
        key = cell_block.type
        data = cell_block.data
        if key in ugrid_counts:
            if ugrid_counts[key] > 0:
                raise ValueError("Ugrid can only handle one cell block of a type.")
            ugrid_counts[key] = data.shape[0]
            ugrid_meshio_id[key] = i
        else:
            msg = f"UGRID mesh format doesn't know {key} cells. Skipping."
            warn(msg)
            continue

    nitems = np.array([list(ugrid_counts.values())])
    # header

    # fortran_header corresponds to the number of bytes in each record
    # it has to be before and after each record
    fortran_header = None
    if file_type["type"] == "F":
        fortran_header = np.array([nitems.nbytes])
        _write_section(f, file_type, fortran_header, itype)

    _write_section(f, file_type, nitems, itype)

    if file_type["type"] == "F":
        # finish current record
        _write_section(f, file_type, fortran_header, itype)

        # start next record
        fortran_header = mesh.points.nbytes
        for cell_block in mesh.cells:
            fortran_header += cell_block.data.nbytes
        # boundary tags
        if ugrid_counts["triangle"] > 0:
            fortran_header += ugrid_counts["triangle"] * np.dtype(itype).itemsize
        if ugrid_counts["quad"] > 0:
            fortran_header += ugrid_counts["quad"] * np.dtype(itype).itemsize
        fortran_header = np.array([fortran_header])
        _write_section(f, file_type, fortran_header, itype)

    _write_section(f, file_type, mesh.points, ftype)

    for key in ["triangle", "quad"]:
        if ugrid_counts[key] == 0:
            continue
        c = mesh.cells[ugrid_meshio_id[key]]
        # UGRID is one-based
        _write_section(f, file_type, c.data + 1, itype)

    # write boundary tags
    for key in ["triangle", "quad"]:
        if ugrid_counts[key] == 0:
            continue

        # pick out cell data
        # for data in mesh.cell_data.values():
        #     if data.dtype in [np.int8, np.int16, np.int32, np.int64]:
        #         labels = data
        #         break

        # pick out cell_data
        labels_key, other = _pick_first_int_data(mesh.cell_data)
        if labels_key and other:
            warn(
                "UGRID can only write one cell data array. "
                f'Picking {labels_key}, skipping {", ".join(other)}.'
            )
        labels = (
            mesh.cell_data[labels_key]
            if labels_key
            else np.ones(ugrid_counts[key], dtype=int)
        )

        labels = labels.reshape(ugrid_counts[key], 1)
        _write_section(f, file_type, labels, itype)

    # write volume elements
    for key in ["tetra", "pyramid", "wedge", "hexahedron"]:
        if ugrid_counts[key] == 0:
            continue
        c = mesh.cells[ugrid_meshio_id[key]]
        # UGRID is one-based
        out = c.data + 1
        if c.type == "pyramid":
            out = out[:, [1, 0, 4, 2, 3]]
        _write_section(f, file_type, out, itype)

    if file_type["type"] == "F":
        _write_section(f, file_type, fortran_header, itype)


register_format("ugrid", [".ugrid"], read, {"ugrid": write})



"""
I/O for Exodus II.

See
<https://src.fedoraproject.org/repo/pkgs/exodusii/922137.pdf/a45d67f4a1a8762bcf66af2ec6eb35f9/922137.pdf>,
in particular Appendix A (page 171, Implementation of EXODUS II with netCDF).
"""
import datetime
import re

import numpy as np

from ..__about__ import __version__
from .._common import warn
from .._exceptions import ReadError
from .._helpers import register_format
from .._mesh import Mesh

exodus_to_meshio_type = {
    "SPHERE": "vertex",
    # curves
    "BEAM": "line",
    "BEAM2": "line",
    "BEAM3": "line3",
    "BAR2": "line",
    # surfaces
    "SHELL": "quad",
    "SHELL4": "quad",
    "SHELL8": "quad8",
    "SHELL9": "quad9",
    "QUAD": "quad",
    "QUAD4": "quad",
    "QUAD5": "quad5",
    "QUAD8": "quad8",
    "QUAD9": "quad9",
    #
    "TRI": "triangle",
    "TRIANGLE": "triangle",
    "TRI3": "triangle",
    "TRI6": "triangle6",
    "TRI7": "triangle7",
    # 'TRISHELL': 'triangle',
    # 'TRISHELL3': 'triangle',
    # 'TRISHELL6': 'triangle6',
    # 'TRISHELL7': 'triangle',
    #
    # volumes
    "HEX": "hexahedron",
    "HEXAHEDRON": "hexahedron",
    "HEX8": "hexahedron",
    "HEX9": "hexahedron9",
    "HEX20": "hexahedron20",
    "HEX27": "hexahedron27",
    #
    "TETRA": "tetra",
    "TETRA4": "tetra4",
    "TET4": "tetra4",
    "TETRA8": "tetra8",
    "TETRA10": "tetra10",
    "TETRA14": "tetra14",
    #
    "PYRAMID": "pyramid",
    "WEDGE": "wedge",
}
meshio_to_exodus_type = {v: k for k, v in exodus_to_meshio_type.items()}


def read(filename):  # noqa: C901
    import netCDF4

    with netCDF4.Dataset(filename) as nc:
        # assert nc.version == np.float32(5.1)
        # assert nc.api_version == np.float32(5.1)
        # assert nc.floating_point_word_size == 8

        # assert b''.join(nc.variables['coor_names'][0]) == b'X'
        # assert b''.join(nc.variables['coor_names'][1]) == b'Y'
        # assert b''.join(nc.variables['coor_names'][2]) == b'Z'

        points = np.zeros((len(nc.dimensions["num_nodes"]), 3))
        point_data_names = []
        cell_data_names = []
        pd = {}
        cd = {}
        cells = []
        ns_names = []
        # eb_names = []
        ns = []
        point_sets = {}
        info = []

        for key, value in nc.variables.items():
            if key == "info_records":
                value.set_auto_mask(False)
                for c in value[:]:
                    try:
                        info += [b"".join(c).decode("UTF-8")]
                    except UnicodeDecodeError:
                        # https://github.com/nschloe/meshio/issues/983
                        pass
            elif key == "qa_records":
                value.set_auto_mask(False)
                for val in value:
                    info += [b"".join(c).decode("UTF-8") for c in val[:]]
            elif key[:7] == "connect":
                meshio_type = exodus_to_meshio_type[value.elem_type.upper()]
                cells.append((meshio_type, value[:] - 1))
            elif key == "coord":
                points = nc.variables["coord"][:].T
            elif key == "coordx":
                points[:, 0] = value[:]
            elif key == "coordy":
                points[:, 1] = value[:]
            elif key == "coordz":
                points[:, 2] = value[:]
            elif key == "name_nod_var":
                value.set_auto_mask(False)
                point_data_names = [b"".join(c).decode("UTF-8") for c in value[:]]
            elif key[:12] == "vals_nod_var":
                idx = 0 if len(key) == 12 else int(key[12:]) - 1
                value.set_auto_mask(False)
                # For now only take the first value
                pd[idx] = value[0]
                if len(value) > 1:
                    warn("Skipping some time data")
            elif key == "name_elem_var":
                value.set_auto_mask(False)
                cell_data_names = [b"".join(c).decode("UTF-8") for c in value[:]]
            elif key[:13] == "vals_elem_var":
                # eb: element block
                m = re.match("vals_elem_var(\\d+)?(?:eb(\\d+))?", key)
                idx = 0 if m.group(1) is None else int(m.group(1)) - 1
                block = 0 if m.group(2) is None else int(m.group(2)) - 1

                value.set_auto_mask(False)
                # For now only take the first value
                if idx not in cd:
                    cd[idx] = {}
                cd[idx][block] = value[0]

                if len(value) > 1:
                    warn("Skipping some time data")
            elif key == "ns_names":
                value.set_auto_mask(False)
                ns_names = [b"".join(c).decode("UTF-8") for c in value[:]]
            # elif key == "eb_names":
            #     value.set_auto_mask(False)
            #     eb_names = [b"".join(c).decode("UTF-8") for c in value[:]]
            elif key.startswith("node_ns"):  # Expected keys: node_ns1, node_ns2
                ns.append(value[:] - 1)  # Exodus is 1-based

        # merge element block data; can't handle blocks yet
        for k, value in cd.items():
            cd[k] = np.concatenate(list(value.values()))

        # Check if there are any <name>R, <name>Z tuples or <name>X, <name>Y, <name>Z
        # triplets in the point data. If yes, they belong together.
        single, double, triple = categorize(point_data_names)

        point_data = {}
        for name, idx in single:
            point_data[name] = pd[idx]
        for name, idx0, idx1 in double:
            point_data[name] = np.column_stack([pd[idx0], pd[idx1]])
        for name, idx0, idx1, idx2 in triple:
            point_data[name] = np.column_stack([pd[idx0], pd[idx1], pd[idx2]])

        cell_data = {}
        k = 0
        for _, cell in cells:
            n = len(cell)
            for name, data in zip(cell_data_names, cd.values()):
                if name not in cell_data:
                    cell_data[name] = []
                cell_data[name].append(data[k : k + n])
            k += n

        point_sets = {name: dat for name, dat in zip(ns_names, ns)}

    return Mesh(
        points,
        cells,
        point_data=point_data,
        cell_data=cell_data,
        point_sets=point_sets,
        info=info,
    )


def categorize(names):
    # Check if there are any <name>R, <name>Z tuples or <name>X, <name>Y, <name>Z
    # triplets in the point data. If yes, they belong together.
    single = []
    double = []
    triple = []
    is_accounted_for = [False] * len(names)
    k = 0
    while True:
        if k == len(names):
            break
        if is_accounted_for[k]:
            k += 1
            continue
        name = names[k]
        if name[-1] == "X":
            ix = k
            try:
                iy = names.index(name[:-1] + "Y")
            except ValueError:
                iy = None
            try:
                iz = names.index(name[:-1] + "Z")
            except ValueError:
                iz = None
            if iy and iz:
                triple.append((name[:-1], ix, iy, iz))
                is_accounted_for[ix] = True
                is_accounted_for[iy] = True
                is_accounted_for[iz] = True
            else:
                single.append((name, ix))
                is_accounted_for[ix] = True
        elif name[-2:] == "_R":
            ir = k
            try:
                iz = names.index(name[:-2] + "_Z")
            except ValueError:
                iz = None
            if iz:
                double.append((name[:-2], ir, iz))
                is_accounted_for[ir] = True
                is_accounted_for[iz] = True
            else:
                single.append((name, ir))
                is_accounted_for[ir] = True
        else:
            single.append((name, k))
            is_accounted_for[k] = True

        k += 1

    if not all(is_accounted_for):
        raise ReadError()
    return single, double, triple


numpy_to_exodus_dtype = {
    "float32": "f4",
    "float64": "f8",
    "int8": "i1",
    "int16": "i2",
    "int32": "i4",
    "int64": "i8",
    "uint8": "u1",
    "uint16": "u2",
    "uint32": "u4",
    "uint64": "u8",
}


def write(filename, mesh):
    import netCDF4

    with netCDF4.Dataset(filename, "w") as rootgrp:
        # set global data
        now = datetime.datetime.now().isoformat()
        rootgrp.title = f"Created by meshio v{__version__}, {now}"
        rootgrp.version = np.float32(5.1)
        rootgrp.api_version = np.float32(5.1)
        rootgrp.floating_point_word_size = 8

        # set dimensions
        total_num_elems = sum(c.data.shape[0] for c in mesh.cells)
        rootgrp.createDimension("num_nodes", len(mesh.points))
        rootgrp.createDimension("num_dim", mesh.points.shape[1])
        rootgrp.createDimension("num_elem", total_num_elems)
        rootgrp.createDimension("num_el_blk", len(mesh.cells))
        rootgrp.createDimension("num_node_sets", len(mesh.point_sets))
        rootgrp.createDimension("len_string", 33)
        rootgrp.createDimension("len_line", 81)
        rootgrp.createDimension("four", 4)
        rootgrp.createDimension("time_step", None)

        # dummy time step
        data = rootgrp.createVariable("time_whole", "f4", ("time_step",))
        data[:] = 0.0

        # points
        coor_names = rootgrp.createVariable(
            "coor_names", "S1", ("num_dim", "len_string")
        )
        coor_names.set_auto_mask(False)
        coor_names[0, 0] = b"X"
        coor_names[1, 0] = b"Y"
        if mesh.points.shape[1] == 3:
            coor_names[2, 0] = b"Z"
        data = rootgrp.createVariable(
            "coord",
            numpy_to_exodus_dtype[mesh.points.dtype.name],
            ("num_dim", "num_nodes"),
        )
        data[:] = mesh.points.T

        # cells
        # ParaView needs eb_prop1 -- some ID. The values don't seem to matter as
        # long as they are different for the for different blocks.
        data = rootgrp.createVariable("eb_prop1", "i4", "num_el_blk")
        for k in range(len(mesh.cells)):
            data[k] = k
        for k, cell_block in enumerate(mesh.cells):
            dim1 = f"num_el_in_blk{k + 1}"
            dim2 = f"num_nod_per_el{k + 1}"
            rootgrp.createDimension(dim1, cell_block.data.shape[0])
            rootgrp.createDimension(dim2, cell_block.data.shape[1])
            dtype = numpy_to_exodus_dtype[cell_block.data.dtype.name]
            data = rootgrp.createVariable(f"connect{k + 1}", dtype, (dim1, dim2))
            data.elem_type = meshio_to_exodus_type[cell_block.type]
            # Exodus is 1-based
            data[:] = cell_block.data + 1

        # point data
        # The variable `name_nod_var` holds the names and indices of the node variables, the
        # variables `vals_nod_var{1,2,...}` hold the actual data.
        num_nod_var = len(mesh.point_data)
        if num_nod_var > 0:
            rootgrp.createDimension("num_nod_var", num_nod_var)
            # set names
            point_data_names = rootgrp.createVariable(
                "name_nod_var", "S1", ("num_nod_var", "len_string")
            )
            point_data_names.set_auto_mask(False)
            for k, name in enumerate(mesh.point_data.keys()):
                for i, letter in enumerate(name):
                    point_data_names[k, i] = letter.encode()

            # Set data. ParaView might have some problems here, see
            # <https://gitlab.kitware.com/paraview/paraview/-/issues/18403>.
            for k, (name, data) in enumerate(mesh.point_data.items()):
                for i, s in enumerate(data.shape):
                    rootgrp.createDimension(f"dim_nod_var{k}{i}", s)
                dims = ["time_step"] + [
                    f"dim_nod_var{k}{i}" for i in range(len(data.shape))
                ]
                node_data = rootgrp.createVariable(
                    f"vals_nod_var{k + 1}",
                    numpy_to_exodus_dtype[data.dtype.name],
                    tuple(dims),
                    fill_value=False,
                )
                node_data[0] = data

        # node sets
        num_point_sets = len(mesh.point_sets)
        if num_point_sets > 0:
            data = rootgrp.createVariable("ns_prop1", "i4", "num_node_sets")
            data_names = rootgrp.createVariable(
                "ns_names", "S1", ("num_node_sets", "len_string")
            )
            for k, name in enumerate(mesh.point_sets.keys()):
                data[k] = k
                for i, letter in enumerate(name):
                    data_names[k, i] = letter.encode()
            for k, (key, values) in enumerate(mesh.point_sets.items()):
                dim1 = f"num_nod_ns{k + 1}"
                rootgrp.createDimension(dim1, values.shape[0])
                dtype = numpy_to_exodus_dtype[values.dtype.name]
                data = rootgrp.createVariable(f"node_ns{k + 1}", dtype, (dim1,))
                # Exodus is 1-based
                data[:] = values + 1


register_format("exodus", [".e", ".exo", ".ex2"], read, {"exodus": write})



import os
import copy

import numpy as np
from scipy import spatial

# from PySide6 import QtCore, QtGui

# import GraphicsItemsCollection as gic
# import GraphicsItem


class Connect:
    """docstring"""

    def __init__(self, progdialog):

        # get MainWindow instance (overcomes handling parents)
        # self.mainwindow = QtCore.QCoreApplication.instance().mainwindow

        self.progdialog = progdialog

    def getVertices(self, block):
        """Make a list of point tuples from a BlockMesh object

        Args:
            block (BlockMesh): BlockMesh object

        Returns:
            list: list of point tuples
                  # [(x1, y1), (x2, y2), (x3, y3), ... , (xn, yn)]
        """
        vertices = list()
        for uline in block.getULines():
            vertices += uline
        return vertices

    def getConnectivity(self, block):

        connectivity = list()

        U, V = block.getDivUV()
        up = U + 1
        for u in range(U):
            for v in range(V):
                p1 = v * up + u
                p2 = p1 + up
                p3 = p2 + 1
                p4 = p1 + 1
                connectivity.append((p1, p2, p3, p4))

        return connectivity

    def getMinMaxConnectivityIDs(self, connectivity):
        id_min = 1e10
        id_max = -1
        for cell in connectivity:
            for id in cell:
                id_min = min(id_min, id)
                id_max = max(id_max, id)
        return id_min, id_max

    def getNearestNeighboursPairs(self, vertices, radius=1.e-6):
        tree = spatial.cKDTree(vertices)
        pairs = tree.query_pairs(radius, p=2., eps=0)
        return pairs

    def getNearestNeighboursBiDirectional(d1, d2, radius=1.e-6):
        """Get all indices ofts in d1 which are within distance r to d2"""
        tree_1 = spatial.cKDTree(d1)
        tree_2 = spatial.cKDTree(d2)
        idx1 = tree_2.query_ball_tree(tree_1, radius, p=2., eps=0)
        idx2 = tree_1.query_ball_tree(tree_2, radius, p=2., eps=0)
        matching = [e[0] for e in idx1 if e]
        opposite = [e[0] for e in idx2 if e]
        return matching, opposite

    def getNearestNeighbours(self, vertices, neighbours, radius=1.e-6):
        """Get the nearest neighbours to each vertex in a list of vertices
        uses Scipy kd-tree for quick nearest-neighbor lookup

        Args:
            vertices (list of tuples): Vertices for which nearest neighbours
                                       should be searched
            neighbours (list of tuples): These are the neighbours which
                                         are being searched
            radius (float, optional): Search neighbours within this radius

        Returns:
            vertex_and_neighbours(dictionary): Contains vertices searched
                                               as key and a list of nearest
                                               neighbours as values
        """

        # setup k-dimensional tree
        tree = spatial.cKDTree(neighbours)

        vertex_and_neighbours = dict()
        for vertex_id, vertex in enumerate(vertices):
            vertex_and_neighbours[vertex_id] = \
                tree.query_ball_point(vertex, radius)

        return vertex_and_neighbours

    def shiftConnectivity(self, connectivity, shift):

        if shift == 0:
            return connectivity

        connectivity_shifted = list()
        for cell in connectivity:
            new_cell = [vertex + shift for vertex in cell]
            connectivity_shifted.append(new_cell)

        return connectivity_shifted

    def connectAllBlocks(self, blocks):

        # compile global vertex list and cell connectivity from all blocks
        vertices = list()
        connectivity = list()

        for i, block in enumerate(blocks):

            # accumulated number of vertices
            # for i = 0 shift is automatically 0
            # so the connectivity of the first block doesn't get shifted
            # thus, this variable must be set before 'vertices += ...'
            shift = len(vertices)

            # concatenate vertices of all blocks
            # vertices += [vertex for vertex in self.getVertices(block)]
            vertices += self.getVertices(block)

            # shift the block connectivity by accumulated number of vertices
            # from all blocks before this one
            connectivity_block = \
                self.shiftConnectivity(self.getConnectivity(block), shift)
            connectivity += [tuple(cell) for cell in connectivity_block]

        if self.progdialog:
            self.progdialog.setValue(80)

        # FIXME
        # FIXME for some reason tuples need to be redefined
        # FIXME
        vertices = [(vertex[0], vertex[1]) for vertex in vertices]

        # search vertices of all blocks against themselves
        # finds itself AND multiple connections (i.e. vertices from neighbour blocks)
        # uses Scipy kd-tree for quick nearest-neighbor lookup
        # the distance tolerance is specified via the radius variable
        vertex_and_neighbours = self.getNearestNeighbours(vertices,
                                                          vertices,
                                                          radius=1.e-6)

        # substitute vertex ids in connectivity at block connections
        connectivity_connected = list()
        for cell in connectivity:
            cell_new = list()
            for node in cell:
                # if there is only one vertex in vertex_and_neighbours,
                # then it is taken as it is
                # if there is more than one vertex,
                # then the minimum vertex index is used
                # so a few vertices remain unused and need to be removed later
                node_new = min(vertex_and_neighbours[node])
                cell_new.append(node_new)
            connectivity_connected.append(cell_new)

        # use numpy arrays
        unconnected = np.array(connectivity)
        connected = np.array(connectivity_connected)

        # deleted nodes
        deleted_nodes = np.unique(unconnected[np.where(connected != unconnected)])

        # delete unused vertices
        vertices_clean = [v for i,v in enumerate(vertices)
                        if i not in sorted(deleted_nodes.tolist())]

        # find remaining node ids
        remaining_nodes = np.setdiff1d(np.unique(connected), deleted_nodes)

        # replace node ids so that a contiguous numbering is established
        # divakar, method 3 (https://stackoverflow.com/a/55950051/2264936)
        mapping = {rn:i for i, rn in enumerate(remaining_nodes)}
        k = np.array(list(mapping.keys()))
        v = np.array(list(mapping.values()))
        mapping_ar = np.zeros(k.max()+1,dtype=v.dtype)
        mapping_ar[k] = v
        connectivity_clean = mapping_ar[connected]

        if self.progdialog:
            self.progdialog.setValue(90)

        # DEBUGGING
        # self.write_debug(unconnected, connected, deleted_nodes, vertices, vertices_clean, connectivity_clean)
        # self.draw_connectivity(vertices, deleted_nodes)

        return (vertices_clean, connectivity_clean, self.progdialog)

    # def draw_connectivity(self, vertices, deleted_nodes):

    #     self.connections = list()

    #     # instantiate a graphics item
    #     marker = gic.GraphicsCollection()
    #      # set its properties
    #     marker.pen.setColor(QtGui.QColor(60, 60, 255, 255))
    #     marker.brush.setColor(QtGui.QColor(255, 50, 50, 230))
    #     marker.pen.setWidthF(1.6)
    #     # no pen thickness change when zoomed
    #     marker.pen.setCosmetic(True)

    #     for node in deleted_nodes:
    #         marker.Circle(vertices[node][0], vertices[node][1], 0.003)
    #         marker_item = GraphicsItem.GraphicsItem(marker)
    #         self.connections.append(marker_item)
            
    #     # add to the scene
    #     self.connections = self.mainwindow.scene. \
    #         createItemGroup(self.connections)

    def write_debug(self, unconnected, connected, deleted_nodes, vertices, vertices_clean, connectivity_clean):

        dicts = locals()
        # print('dicts keys', dicts.keys())
        # print('type', type(dicts))
        dicts.pop('self')

        folder = 'debug'
        if not os.path.isdir(folder):
            os.mkdir(folder)

        # write all data to individual files
        for listname in dicts:
            with open(os.path.join(folder, listname + '.txt'), 'w') as f:
                for item in dicts[listname]:
                    f.write(str(item) + '\n')



from __future__ import annotations

from xml.etree import ElementTree as ET

import numpy as np
from rich.console import Console

# See <https://github.com/nschloe/meshio/wiki/Node-ordering-in-cells> for the node
# ordering.
num_nodes_per_cell = {
    "vertex": 1,
    "line": 2,
    "triangle": 3,
    "quad": 4,
    "quad8": 8,
    "tetra": 4,
    "hexahedron": 8,
    "hexahedron20": 20,
    "hexahedron24": 24,
    "wedge": 6,
    "pyramid": 5,
    #
    "line3": 3,
    "triangle6": 6,
    "quad9": 9,
    "tetra10": 10,
    "hexahedron27": 27,
    "wedge15": 15,
    "wedge18": 18,
    "pyramid13": 13,
    "pyramid14": 14,
    #
    "line4": 4,
    "triangle10": 10,
    "quad16": 16,
    "tetra20": 20,
    "wedge40": 40,
    "hexahedron64": 64,
    #
    "line5": 5,
    "triangle15": 15,
    "quad25": 25,
    "tetra35": 35,
    "wedge75": 75,
    "hexahedron125": 125,
    #
    "line6": 6,
    "triangle21": 21,
    "quad36": 36,
    "tetra56": 56,
    "wedge126": 126,
    "hexahedron216": 216,
    #
    "line7": 7,
    "triangle28": 28,
    "quad49": 49,
    "tetra84": 84,
    "wedge196": 196,
    "hexahedron343": 343,
    #
    "line8": 8,
    "triangle36": 36,
    "quad64": 64,
    "tetra120": 120,
    "wedge288": 288,
    "hexahedron512": 512,
    #
    "line9": 9,
    "triangle45": 45,
    "quad81": 81,
    "tetra165": 165,
    "wedge405": 405,
    "hexahedron729": 729,
    #
    "line10": 10,
    "triangle55": 55,
    "quad100": 100,
    "tetra220": 220,
    "wedge550": 550,
    "hexahedron1000": 1000,
    "hexahedron1331": 1331,
    #
    "line11": 11,
    "triangle66": 66,
    "quad121": 121,
    "tetra286": 286,
}


def cell_data_from_raw(cells, cell_data_raw):
    cs = np.cumsum([len(c) for c in cells])[:-1]
    return {name: np.split(d, cs) for name, d in cell_data_raw.items()}


def raw_from_cell_data(cell_data):
    return {name: np.concatenate(value) for name, value in cell_data.items()}


def write_xml(filename, root):
    tree = ET.ElementTree(root)
    tree.write(filename)


def _pick_first_int_data(data):
    # pick out material
    keys = list(data.keys())
    candidate_keys = []
    for key in keys:
        # works for point_data and cell_data
        if data[key][0].dtype.kind in ["i", "u"]:  # int or uint
            candidate_keys.append(key)

    if len(candidate_keys) > 0:
        # pick the first
        key = candidate_keys[0]
        idx = keys.index(key)
        other = keys[:idx] + keys[idx + 1 :]
    else:
        key = None
        other = []

    return key, other


def info(string, highlight: bool = True) -> None:
    Console(stderr=True).print(f"[bold]Info:[/bold] {string}", highlight=highlight)


def warn(string, highlight: bool = True) -> None:
    Console(stderr=True).print(
        f"[yellow][bold]Warning:[/bold] {string}[/yellow]", highlight=highlight
    )


def error(string, highlight: bool = True) -> None:
    Console(stderr=True).print(
        f"[red][bold]Error:[/bold] {string}[/red]", highlight=highlight
    )


def is_in_any(string: str, strings: list[str]) -> bool:
    """True if `string` is contained in any of `strings`."""
    for s in strings:
        if string in s:
            return True
    return False


def join_strings(strings: list[str]) -> tuple[str, str]:
    """Join strings such that they can be uniquely split again afterwards."""
    possible_join_chars = ["-", "_", "#", "+", "/"]
    char = None
    for c in possible_join_chars:
        if not is_in_any(c, strings):
            char = c
            break
    assert char is not None
    return char.join(strings), char


def replace_space(string: str) -> tuple[str, str]:
    possible_chars = ["_", "-", "+", "X", "/", "#"]
    char = None
    for c in possible_chars:
        if c not in string:
            char = c
            break
    assert char is not None
    return string.replace(" ", char), char



from setuptools import setup,find_packages
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='FoilMesh',
    author='Zcaic',
    version='0.0.8',
    description='FoilMesh is a tool to mesh airfoil structural grid',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/Zcaic/Foilmesh",
    keywords="aerodynamics cfd airfoil grid mesh",
    packages=find_packages(),
    python_requires='>=3.0',
    install_requires=[
        'numpy',
        'scipy',
        'rich'              
        ],
)


import re
from collections import OrderedDict
from io import StringIO

import numpy as np

from .._common import warn
from .._exceptions import ReadError
from .._files import open_file
from .._helpers import register_format
from .._mesh import CellBlock, Mesh

float_pattern = r"[+-]?(?:\d+\.?\d*|\d*\.?\d+)"
float_re = re.compile(float_pattern)

point_pattern = r"{0}\s+{0}\s+{0}(?:\s+{0})?".format(float_pattern)
point_re = re.compile(point_pattern)

triangle_pattern = r"\(\s*\(\s*({})\s*\)\s*\)".format(
    r"\s*,\s*".join(point_pattern for _ in range(4))
)
triangle_re = re.compile(triangle_pattern)

tin_pattern = rf"TIN\s*\((?:\s*{triangle_pattern}\s*,?)*\s*\)"
tin_re = re.compile(tin_pattern)


def read_str(s):
    s = s.strip()
    tin_match = tin_re.match(s)
    if tin_match is None:
        raise ReadError("Invalid WKT TIN")
    point_idxs = OrderedDict()
    tri_idxs = []
    for tri_match in triangle_re.finditer(tin_match.group()):
        tri_point_idxs = []
        for point_match in point_re.finditer(tri_match.group()):
            point = []
            for float_match in float_re.finditer(point_match.group()):
                point.append(float(float_match.group()))
            point = tuple(point)
            if point not in point_idxs:
                point_idxs[point] = len(point_idxs)
            tri_point_idxs.append(point_idxs[point])

        if tri_point_idxs[-1] != tri_point_idxs[0]:
            raise ValueError("Triangle is not a closed linestring")

        tri_idxs.append(tri_point_idxs[:-1])

    try:
        point_arr = np.array(list(point_idxs), np.float64)
    except ValueError as e:
        if len({len(p) for p in point_idxs}) > 1:
            raise ReadError("Points have mixed dimensionality")
        else:
            raise e

    tri_arr = np.array(tri_idxs, np.uint64)

    return Mesh(point_arr, [CellBlock("triangle", tri_arr)])


def arr_to_str(arr):
    return " ".join(str(item) for item in arr)


def read(filename):
    with open_file(filename) as f:
        return read_str(f.read())


def write(filename, mesh):
    with open_file(filename, "w") as f:
        write_buffer(f, mesh)


def write_buffer(f, mesh):
    skip = [c for c in mesh.cells if c.type != "triangle"]
    if skip:
        warn('WTK only supports triangle cells. Skipping {", ".join(skip)}.')

    triangles = mesh.get_cells_type("triangle")

    f.write("TIN (")
    joiner = ""
    for tri_points in mesh.points[triangles]:
        f.write(
            "{0}(({1}, {2}, {3}, {1}))".format(
                joiner, *(arr_to_str(p) for p in tri_points)
            )
        )
        joiner = ", "
    f.write(")")


def write_str(mesh):
    buf = StringIO()
    write_buffer(buf, mesh)
    buf.seek(0)
    return buf.read()


register_format("wkt", [".wkt"], read, {"wkt": write})



"""
I/O for KratosMultiphysics's mdpa format, cf.
<https://github.com/KratosMultiphysics/Kratos/wiki/Input-data>.

The MDPA format is unsuitable for fast consumption, this is why:
<https://github.com/KratosMultiphysics/Kratos/issues/5365>.
"""
import numpy as np

from .._common import num_nodes_per_cell, raw_from_cell_data, warn
from .._exceptions import ReadError, WriteError
from .._files import open_file
from .._helpers import register_format
from .._mesh import Mesh

## We check if we can read/write the mesh natively from Kratos
# TODO: Implement native reading

# Translate meshio types to KratosMultiphysics codes
# Kratos uses the same node numbering of GiD pre and post processor
# http://www-opale.inrialpes.fr/Aerochina/info/en/html-version/gid_11.html
# https://github.com/KratosMultiphysics/Kratos/wiki/Mesh-node-ordering
_mdpa_to_meshio_type = {
    "Line2D2": "line",
    "Line3D2": "line",
    "Triangle2D3": "triangle",
    "Triangle3D3": "triangle",
    "Quadrilateral2D4": "quad",
    "Quadrilateral3D4": "quad",
    "Tetrahedra3D4": "tetra",
    "Hexahedra3D8": "hexahedron",
    "Prism3D6": "wedge",
    "Line2D3": "line3",
    "Triangle2D6": "triangle6",
    "Triangle3D6": "triangle6",
    "Quadrilateral2D9": "quad9",
    "Quadrilateral3D9": "quad9",
    "Tetrahedra3D10": "tetra10",
    "Hexahedra3D27": "hexahedron27",
    "Point2D": "vertex",
    "Point3D": "vertex",
    "Quadrilateral2D8": "quad8",
    "Quadrilateral3D8": "quad8",
    "Hexahedra3D20": "hexahedron20",
}

_meshio_to_mdpa_type = {
    "line": "Line2D2",
    "triangle": "Triangle2D3",
    "quad": "Quadrilateral2D4",
    "tetra": "Tetrahedra3D4",
    "hexahedron": "Hexahedra3D8",
    "wedge": "Prism3D6",
    "line3": "Line2D3",
    "triangle6": "Triangle2D6",
    "quad9": "Quadrilateral2D9",
    "tetra10": "Tetrahedra3D10",
    "hexahedron27": "Hexahedra3D27",
    "vertex": "Point2D",
    "quad8": "Quadrilateral2D8",
    "hexahedron20": "Hexahedra3D20",
}
inverse_num_nodes_per_cell = {v: k for k, v in num_nodes_per_cell.items()}

local_dimension_types = {
    "Line2D2": 1,
    "Line3D2": 1,
    "Triangle2D3": 2,
    "Triangle3D3": 2,
    "Quadrilateral2D4": 2,
    "Quadrilateral3D4": 2,
    "Tetrahedra3D4": 3,
    "Hexahedra3D8": 3,
    "Prism3D6": 3,
    "Line2D3": 1,
    "Triangle2D6": 2,
    "Triangle3D6": 2,
    "Quadrilateral2D9": 2,
    "Quadrilateral3D9": 2,
    "Tetrahedra3D10": 3,
    "Hexahedra3D27": 3,
    "Point2D": 0,
    "Point3D": 0,
    "Quadrilateral2D8": 2,
    "Quadrilateral3D8": 2,
    "Hexahedra3D20": 3,
}


def read(filename):
    """Reads a KratosMultiphysics mdpa file."""
    # if (have_kratos is True): # TODO: Implement natively
    # pass
    # else:
    with open_file(filename, "rb") as f:
        mesh = read_buffer(f)
    return mesh


def _read_nodes(f, is_ascii, data_size):
    # Count the number of nodes. This is _extremely_ ugly; we first read the _entire_
    # file until "End Nodes". The crazy thing is that first counting the lines, then
    # skipping back to pos, and using fromfile there is _faster_ than accumulating the
    # points into a list and converting them to a numpy array afterwards. A point count
    # would be _really_ helpful here, but yeah, that's a fallacy of the format.
    # <https://github.com/KratosMultiphysics/Kratos/issues/5365>
    pos = f.tell()
    num_nodes = 0
    while True:
        line = f.readline().decode()
        if "End Nodes" in line:
            break
        num_nodes += 1
    f.seek(pos)

    points = np.fromfile(f, count=num_nodes * 4, sep=" ").reshape((num_nodes, 4))
    # The first number is the index
    points = points[:, 1:]

    line = f.readline().decode()
    if line.strip() != "End Nodes":
        raise ReadError()
    return points


def _read_cells(f, cells, is_ascii, cell_tags, environ=None):
    if not is_ascii:
        raise ReadError("Can only read ASCII cells")

    # First we try to identify the entity
    t = None
    if environ is not None:
        if environ.startswith("Begin Elements "):
            entity_name = environ[15:]
            for key in _mdpa_to_meshio_type:
                if key in entity_name:
                    t = _mdpa_to_meshio_type[key]
                    break
        elif environ.startswith("Begin Conditions "):
            entity_name = environ[17:]
            for key in _mdpa_to_meshio_type:
                if key in entity_name:
                    t = _mdpa_to_meshio_type[key]
                    break

    while True:
        line = f.readline().decode()
        if line.startswith("End Elements") or line.startswith("End Conditions"):
            break
        # data[0] gives the entity id
        # data[1] gives the property id
        # The rest are the ids of the nodes
        data = [int(k) for k in filter(None, line.split())]
        num_nodes_per_elem = len(data) - 2
        # We use this in case not alternative
        if t is None:
            t = inverse_num_nodes_per_cell[num_nodes_per_elem]

        if len(cells) == 0 or t != cells[-1][0]:
            cells.append((t, []))
        # Subtract one to account for the fact that python indices are 0-based.
        cells[-1][1].append(np.array(data[-num_nodes_per_elem:]) - 1)

        # Using the property id as tag
        if t not in cell_tags:
            cell_tags[t] = []
        cell_tags[t].append([data[1]])

    # Cannot convert cell_tags[key] to numpy array: There may be a
    # different number of tags for each cell.

    if line.strip() not in ["End Elements", "End Conditions"]:
        raise ReadError()


def _prepare_cells(cells, cell_tags):
    # Declaring has additional data tag
    has_additional_tag_data = False

    # restrict to the standard two data items (physical, geometrical)
    output_cell_tags = {}
    for key in cell_tags:
        output_cell_tags[key] = {"gmsh:physical": [], "gmsh:geometrical": []}
        for item in cell_tags[key]:
            if len(item) > 0:
                output_cell_tags[key]["gmsh:physical"].append(item[0])
            if len(item) > 1:
                output_cell_tags[key]["gmsh:geometrical"].append(item[1])
            if len(item) > 2:
                has_additional_tag_data = True
        output_cell_tags[key]["gmsh:physical"] = np.array(
            output_cell_tags[key]["gmsh:physical"], dtype=int
        )
        output_cell_tags[key]["gmsh:geometrical"] = np.array(
            output_cell_tags[key]["gmsh:geometrical"], dtype=int
        )

    # Kratos cells are mostly ordered like VTK, with a few exceptions:
    if "hexahedron20" in cells:
        cells["hexahedron20"] = cells["hexahedron20"][
            :, [0, 1, 2, 3, 4, 5, 6, 7, 8, 11, 10, 9, 16, 19, 18, 17, 12, 13, 14, 15]
        ]
    if "hexahedron27" in cells:
        cells["hexahedron27"] = cells["hexahedron27"][
            :,
            [
                0,
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                11,
                10,
                9,
                16,
                19,
                18,
                17,
                12,
                13,
                14,
                15,
                22,
                24,
                21,
                23,
                20,
                25,
                26,
            ],
        ]

    return has_additional_tag_data


# def _read_data(f, tag, data_dict, is_ascii):
#     if not is_ascii:
#         raise ReadError()
#     # Read string tags
#     num_string_tags = int(f.readline().decode())
#     string_tags = [
#         f.readline().decode().strip().replace('"', "")
#         for _ in range(num_string_tags)
#     ]
#     # The real tags typically only contain one value, the time.
#     # Discard it.
#     num_real_tags = int(f.readline().decode())
#     for _ in range(num_real_tags):
#         f.readline()
#     num_integer_tags = int(f.readline().decode())
#     integer_tags = [int(f.readline().decode()) for _ in range(num_integer_tags)]
#     num_components = integer_tags[1]
#     num_items = integer_tags[2]
#
#     # Creating data
#     data = np.fromfile(f, count=num_items * (1 + num_components), sep=" ").reshape(
#         (num_items, 1 + num_components)
#     )
#     # The first number is the index
#     data = data[:, 1:]
#
#     line = f.readline().decode()
#     if line.strip() != f"End {tag}":
#         raise ReadError()
#
#     # The gmsh format cannot distinguish between data of shape (n,) and (n, 1).
#     # If shape[1] == 1, cut it off.
#     if data.shape[1] == 1:
#         data = data[:, 0]
#
#     data_dict[string_tags[0]] = data


def read_buffer(f):
    # The format is specified at
    # <https://github.com/KratosMultiphysics/Kratos/wiki/Input-data>.

    # Initialize the optional data fields
    points = []
    cells = []
    field_data = {}
    cell_data = {}
    # cell_data_raw = {}
    cell_tags = {}
    point_data = {}

    is_ascii = True
    data_size = None

    # Definition of cell tags
    cell_tags = {}

    # Saving position
    # pos = f.tell()
    # Read mesh
    while True:
        line = f.readline().decode()
        if not line:
            # EOF
            break
        environ = line.strip()

        if environ.startswith("Begin Nodes"):
            points = _read_nodes(f, is_ascii, data_size)
        elif environ.startswith("Begin Elements") or environ.startswith(
            "Begin Conditions"
        ):
            _read_cells(f, cells, is_ascii, cell_tags, environ)

    # We finally prepare the cells
    has_additional_tag_data = _prepare_cells(cells, cell_tags)

    # Reverting to the original position
    # f.seek(pos)
    # Read data
    # TODO: To implement
    # while False:
    #     line = f.readline().decode()
    #     if not line:
    #         # EOF
    #         break
    #     # elif "NodalData" in environ and cells_prepared:
    #     # _read_data(f, "NodalData", point_data, data_size, is_ascii)
    #     # elif "Begin ElementalData" in environ:
    #     # _read_data(f, "ElementalData", cell_data_raw, data_size, is_ascii)
    #     # elif "Begin ConditionalData" in environ:
    #     # _read_data(f, "ConditionalData", cell_data_raw, data_size, is_ascii)

    if has_additional_tag_data:
        warn("The file contains tag data that couldn't be processed.")

    # cell_data = cell_data_from_raw(cells, cell_data_raw)

    ## Merge cell_tags into cell_data
    # for key, tag_dict in cell_tags.items():
    # if key not in cell_data:
    # cell_data[key] = {}
    # for name, item_list in tag_dict.items():
    # assert name not in cell_data[key]
    # cell_data[key][name] = item_list

    return Mesh(
        points, cells, point_data=point_data, cell_data=cell_data, field_data=field_data
    )


def cell_data_from_raw(cells, cell_data_raw):
    cell_data = {k: {} for k in cells}
    for key in cell_data_raw:
        d = cell_data_raw[key]
        r = 0
        for k in cells:
            cell_data[k][key] = d[r : r + len(cells[k])]
            r += len(cells[k])

    return cell_data


def _write_nodes(fh, points, float_fmt, binary=False):
    fh.write(b"Begin Nodes\n")
    if binary:
        raise WriteError()

    for k, x in enumerate(points):
        fmt = " {} " + " ".join(3 * ["{:" + float_fmt + "}"]) + "\n"
        fh.write(fmt.format(k + 1, x[0], x[1], x[2]).encode())
    fh.write(b"End Nodes\n\n")


def _write_elements_and_conditions(fh, cells, tag_data, binary=False, dimension=2):
    if binary:
        raise WriteError("Can only write ASCII")
    # write elements
    entity = "Elements"
    dimension_name = f"{dimension}D"
    wrong_dimension_name = "3D" if dimension == 2 else "2D"
    consecutive_index = 0
    for cell_block in cells:
        cell_type = cell_block.type
        node_idcs = cell_block.data
        # NOTE: The names of the dummy conditions are not regular, require extra work
        # local_dimension = local_dimension_types[cell_type]
        # if (local_dimension < dimension):
        # entity = "Conditions"
        mdpa_cell_type = _meshio_to_mdpa_type[cell_type].replace(
            wrong_dimension_name, dimension_name
        )
        fh.write(f"Begin {entity} {mdpa_cell_type}\n".encode())

        # TODO: Add proper tag recognition in the future
        fcd = np.empty((len(node_idcs), 0), dtype=np.int32)
        for k, c in enumerate(node_idcs):
            a1 = " ".join([str(val) for val in fcd[k]])
            a2 = " ".join([str(cc) for cc in c + 1])
            fh.write(
                f" {consecutive_index + k + 1} {fcd.shape[1]} {a1} {a2}\n".encode()
            )

        consecutive_index += len(node_idcs)
        fh.write(f"End {entity}\n\n".encode())


def _write_data(fh, tag, name, data, binary):
    if binary:
        raise WriteError()
    fh.write(f"Begin {tag} {name}\n\n".encode())
    # number of components
    num_components = data.shape[1] if len(data.shape) > 1 else 1

    # Cut off the last dimension in case it's 1. This avoids problems with
    # writing the data.
    if len(data.shape) > 1 and data.shape[1] == 1:
        data = data[:, 0]

    # Actually write the data
    fmt = " ".join(["{}"] + ["{!r}"] * num_components) + "\n"
    # TODO unify
    if num_components == 1:
        for k, x in enumerate(data):
            fh.write(fmt.format(k + 1, x).encode())
    else:
        for k, x in enumerate(data):
            fh.write(fmt.format(k + 1, *x).encode())

    fh.write(f"End {tag} {name}\n\n".encode())


def write(filename, mesh, float_fmt=".16e", binary=False):
    """Writes mdpa files, cf.
    <https://github.com/KratosMultiphysics/Kratos/wiki/Input-data>.
    """
    if binary:
        raise WriteError()
    if mesh.points.shape[1] == 2:
        warn(
            "mdpa requires 3D points, but 2D points given. "
            "Appending 0 third component."
        )
        points = np.column_stack([mesh.points, np.zeros_like(mesh.points[:, 0])])
    else:
        points = mesh.points

    # Kratos cells are mostly ordered like VTK, with a few exceptions:
    cells = mesh.cells.copy()
    if "hexahedron20" in cells:
        cells["hexahedron20"] = cells["hexahedron20"][
            :, [0, 1, 2, 3, 4, 5, 6, 7, 8, 11, 10, 9, 16, 17, 18, 19, 12, 15, 14, 13]
        ]
    if "hexahedron27" in cells:
        cells["hexahedron27"] = cells["hexahedron27"][
            :,
            [
                0,
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                11,
                10,
                9,
                16,
                17,
                18,
                19,
                12,
                15,
                14,
                13,
                22,
                24,
                21,
                23,
                20,
                25,
                26,
            ],
        ]

    with open_file(filename, "wb") as fh:
        # Write some additional info
        fh.write(b"Begin ModelPartData\n")
        fh.write(b"//  VARIABLE_NAME value\n")
        fh.write(b"End ModelPartData\n\n")
        fh.write(b"Begin Properties 0\n")
        fh.write(b"End Properties\n\n")

        # Split the cell data: gmsh:physical and gmsh:geometrical are tags, the
        # rest is actual cell data.
        tag_data = {}
        other_data = {}
        for key, data in mesh.cell_data.items():
            if key in ["gmsh:physical", "gmsh:geometrical"]:
                tag_data[key] = [entry.astype(np.int32) for entry in data]
            else:
                other_data[key] = data

        # identity dimension
        dimension = 2
        for c in cells:
            name_elem = _meshio_to_mdpa_type[c.type]
            if local_dimension_types[name_elem] == 3:
                dimension = 3
                break

        # identify entities
        _write_nodes(fh, points, float_fmt, binary)
        _write_elements_and_conditions(fh, cells, tag_data, binary, dimension)
        for name, dat in mesh.point_data.items():
            _write_data(fh, "NodalData", name, dat, binary)
        cell_data_raw = raw_from_cell_data(other_data)
        for name, dat in cell_data_raw.items():
            # assume always that the components are elements (for now)
            _write_data(fh, "ElementalData", name, dat, binary)


register_format("mdpa", [".mdpa"], read, {"mdpa": write})
